# -*- coding: utf-8 -*-

from typing import Union

import six

from .db import get_session
from .models import Element


__all__ = [
    "get_all_elements",
    "element",
]


def element(ids: Union[int, str]) -> Element:
    """
    Based on the type of the `ids` identifier return either an
    :py:class:`Element <mendeleev.models.Element>` object from the
    database, or a list of :py:class:`Element <mendeleev.models.Element>`
    objects if the `ids` is a list or a tuple of identifiers. Valid
    identifiers for an element are: *name*, *symbol*, and
    *atomic number*.

    Args:
        ids (str): element identifier

    Raises:
        ValueError: when the identifier is not a list/tuple, int or str

    Example:
        The element can be identified by symbol

        >>> from mendeleev import element
        >>> si = element('Si')
        >>> si.atomic_number
        14

        by the atomic number

        >>> al = element(13)
        >>> al.name
        'Aluminum'

        or by the name

        >>> o = element('Oxygen')
        >>> o.symbol
        'O'

        Mutiple elements can be instantiated simultaneously through as
        combination of identifiers

        >>> c, h, o = element(['C', 'Hydrogen', 8])
        >>> print(c.name, h.name, o.name)
        Carbon Hydrogen Oxygen

    """

    if isinstance(ids, (list, tuple)):
        return [_get_element(i) for i in ids]
    elif isinstance(ids, (six.string_types, int)):
        return _get_element(ids)
    else:
        raise ValueError(
            "Expected a <list>, <tuple>, <str> or <int>, got: {0:s}".format(type(ids))
        )


def _get_element(ids):
    """
    Return an element from the database based on the `ids` identifier passed.
    Valid identifiers for an element are: *name*, *symbol*, *atomic number*.
    """

    session = get_session()

    if isinstance(ids, six.string_types):
        if len(ids) <= 3 and ids.lower() != "tin":
            return session.query(Element).filter(Element.symbol == str(ids)).one()
        else:
            return session.query(Element).filter(Element.name == str(ids)).one()
    elif isinstance(ids, int):
        return session.query(Element).filter(Element.atomic_number == ids).one()
    else:
        raise ValueError("Expecting a <str> or <int>, got: {0:s}".format(type(ids)))


def get_all_elements():
    "Get all elements as a list"

    session = get_session()
    elements = session.query(Element).all()
    session.close()
    return elements


def ids_to_attr(ids, attr="atomic_number"):
    """
    Convert the element ids: atomic numbers, symbols, element names or a
    combination of the above to a list of corresponding attributes.

    Args:
      ids: list, str or int
        A list of atomic number, symbols, element names of a combination of
        them
      attr: str
        Name of the desired attribute

    Returns:
      out: list
        List of attributes corresponding to the ids
    """

    if isinstance(ids, (list, tuple)):
        return [getattr(e, attr) for e in element(ids)]
    else:
        return [getattr(element(ids), attr)]


def deltaN(id1, id2, charge1=0, charge2=0, missingIsZero=True):
    """
    Calculate the approximate fraction of transferred electrons between
    elements or ions `id1` and `id2` with charges `charge1` and `charge2`
    respectively according to the expression

    .. math::

       \Delta N = \\frac{\chi_{A} - \chi_{B}}{2(\eta_{A} + \eta_{B})}

    Args:
      id1: str or int
        Element identifier atomic number, symbol or element name
      id2: str or int
        Element identifier atomic number, symbol or element name
    """

    session = get_session()
    atns = ids_to_attr([id1, id2], attr="atomic_number")

    e1, e2 = [
        session.query(Element).filter(Element.atomic_number == a).one() for a in atns
    ]

    chi = [
        x.en_mulliken(charge=c, missingIsZero=missingIsZero)
        for x, c in zip([e1, e2], [charge1, charge2])
    ]

    if all(x is not None for x in chi):
        return (chi[0] - chi[1]) / (
            2.0 * (e1.hardness(charge=charge1) + e2.hardness(charge=charge2))
        )
    else:
        return None



from typing import Tuple
import seaborn as sns
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


def heatmap(
    elements: pd.DataFrame,
    prop: str,
    style: str = "whitegrid",
    figsize: Tuple[int] = (16, 10),
    cmap: str = "RdBu_r",
    lw: int = 1,
    output: str = None,
    **kwargs
):
    """
    Plot a heatmap of the given property

    Args:
        elements: DataFrame with data about elements
        prop : Name of the attribute of Element object that is available from the
            elements table
        style : Seaborn style option, default='whitegrid
        figsize : Size of the plot, default=(16, 10)
        cmap : Colormap to use, default='RdBu_r'
        lw : Seaborn heatmap `linewidths` argumentm default=1,
            see http://stanford.edu/~mwaskom/software/seaborn/generated/seaborn.heatmap.html
        output : File name to save the plot, by default nothing is saved
    """

    # add lanthanides and actinides

    keys = ["period", "group_id", prop]
    els = elements[keys].dropna()
    elements_rect = els.pivot(*keys)

    sns.set(font_scale=1.5, style=style, rc={"figure.figsize": figsize})
    mask = np.asarray(elements_rect.isnull())
    ax = sns.heatmap(elements_rect, cmap=cmap, mask=mask, linewidths=lw, **kwargs)
    n = len(ax.xaxis.get_ticklabels())
    ax.set_yticklabels(elements_rect.index[::-1], rotation=0)
    ax.set_xticklabels(list(range(1, n + 1)))
    ax.xaxis.tick_top()
    ax.xaxis.set_label_position("top")
    ax.set_xlabel("Group")
    ax.set_ylabel("Period")
    if output is not None:
        plt.savefig(output)
    return ax



import urllib
import numpy as np
import pandas as pd
from bs4 import BeautifulSoup


def parse_rowspan_table(table):
    """
    based on http://stackoverflow.com/a/28766684
    """

    tmp = table.find_all("tr")

    first = tmp[0]
    allRows = tmp[1:-1]

    headers = [header.get_text() for header in first.find_all("th")]

    results = [[data.get_text() for data in row.find_all("td")] for row in allRows]

    rowspan = []

    for no, tr in enumerate(allRows):
        for td_no, data in enumerate(tr.find_all("td")):
            if data.has_attr("rowspan"):
                rowspan.append((no, td_no, int(data["rowspan"]), data.get_text()))

    if rowspan:
        for i in rowspan:
            # tr value of rowspan in present in 1th place in results
            for j in range(1, i[2]):
                # Add value in next tr.
                results[i[0] + j].insert(i[1], i[3])

    return pd.DataFrame(data=results, columns=headers)


def ciaaw_atomic_masses():
    """
    Scrape the CIAAW webpage for atomic masses and return them as
    pandas DataFrame
    """

    url = "http://ciaaw.org/atomic-masses.htm"

    req = urllib.request.Request(url)
    page = urllib.request.urlopen(req)
    soup = BeautifulSoup(page, "html5lib")

    table = soup.find_all("table", attrs={"id": "mytable"})[0]
    df = parse_rowspan_table(table)

    df.loc[:, "Z"] = df["Z"].str.strip().astype(int)
    df.loc[:, "Element"] = df["Element"].str.strip()
    df.loc[:, "Symbol"] = df["Symbol"].str.strip("\t").astype(str)
    df.loc[:, "Atomic mass,ma/Da"] = df["Atomic mass,ma/Da"].str.strip()

    df.loc[:, "is_radioactive"] = df["A"].str.contains("*", regex=False)
    df.loc[:, "A"] = df["A"].str.extract("([0-9]+)", expand=False).astype(int)

    df.loc[:, "AM wows"] = df["Atomic mass,ma/Da"].str.replace("\s", "")
    df[["AM lead", "AM decimals", "AM unc"]] = df["AM wows"].str.extract(
        r"(\d+)\.(\d+)\((\d+)\)", expand=True
    )
    df["Atomic Mass"] = df["AM wows"].str.extract(r"(\d+\.\d+)", expand=False)

    # fill in the data for carbon
    df.loc[df["AM unc"].isnull(), "AM lead"] = 12
    df.loc[df["AM unc"].isnull(), "AM unc"] = 0

    # calculate the uncertainty
    mask = df["AM decimals"].notnull()
    dec = np.power(10, -1 * df.loc[mask, "AM decimals"].str.len().astype(float))
    df.loc[mask, "AM uncertainty"] = dec * df.loc[mask, "AM unc"].astype(int)

    df.loc[(df["Symbol"] == "C") & (df["A"] == 12), "Atomic Mass"] = 12.0e0
    df.loc[(df["Symbol"] == "C") & (df["A"] == 12), "AM uncertainty"] = 0.0e0
    df.loc[:, "Atomic Mass"] = df["Atomic Mass"].astype(float)

    df.drop("AM wows", axis=1, inplace=True)
    df.drop("AM lead", axis=1, inplace=True)
    df.drop("AM decimals", axis=1, inplace=True)
    df.drop("AM unc", axis=1, inplace=True)

    return df


def ciaaw_atomic_weights():
    """
    Scrape the CIAAW webpage for atomic weights and return them as
    pandas DataFrame
    """

    url = "http://ciaaw.org/atomic-weights.htm"

    table = pd.read_html(url, attrs={"id": "mytable"})[0]
    # drop the last row
    table = table.iloc[:-1, :]

    table = table.rename(columns={"Standard Atomic Weight": "SAW"})
    table.loc[:, "Z"] = table["Z"].astype(int)
    table.loc[:, "SAW"] = table["SAW"].str.replace("\s", "")
    table[["SAW decimals", "SAW unc"]] = table["SAW"].str.extract(
        r"\d+\.(\d+)\((\d+)\)", expand=True
    )

    # calculate the uncertainty
    mask = table["SAW decimals"].notnull()
    dec = np.power(10, -1 * table.loc[mask, "SAW decimals"].str.len().astype(float))
    table.loc[mask, "SAW uncertainty"] = dec * table.loc[mask, "SAW unc"].astype(int)

    table["Atomic Weight"] = (
        table["SAW"].str.extract(r"(\d+\.\d+)\(", expand=False).astype(float)
    )

    table.drop("SAW decimals", axis=1, inplace=True)
    table.drop("SAW unc", axis=1, inplace=True)

    return table


def ciaaw_isotopic_abundances():

    url = "http://ciaaw.org/isotopic-abundances.htm"

    req = urllib.request.Request(url)
    page = urllib.request.urlopen(req)
    soup = BeautifulSoup(page, "html5lib")

    table = soup.find_all("table", attrs={"id": "mytable"})[0]
    ia = parse_rowspan_table(table)

    abu = "Representative isotopic composition"
    ia.loc[:, "A"] = ia["A"].str.extract("([0-9]+)", expand=False).astype(int)
    ia.loc[:, "Z"] = ia["Z"].astype(int)
    ia.loc[:, abu] = ia[abu].str.replace("\s", "")
    ia.loc[:, "abundance"] = ia[abu].str.extract(r"(\d+\.\d+)\(", expand=False)
    ia.loc[:, "abundance"].fillna(
        ia[abu].str.extract(r"\[(\d+\.\d+),", expand=False), inplace=True
    )

    return ia


def ciaaw_monoisotopic():
    """
    Scrape the CIAAW webpage for monoisotopic elements and return the
    data as pandas DataFrame
    """

    url = "http://ciaaw.org/monoisotopic-elements.htm"

    # parse monoisotopic elements
    table = pd.read_html(url, attrs={"id": "mytable"})[0]
    # drop the last row
    table = table.iloc[:-1, :]
    # convert atomic number to integer
    table.loc[:, "Z"] = table.loc[:, "Z"].astype(int)
    return table



import os

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker


DBNAME = "elements.db"


def get_package_dbpath():
    """Return the default database path"""

    return os.path.join(os.path.abspath(os.path.dirname(__file__)), DBNAME)


def get_engine(dbpath=None):
    """Return the db engine"""

    if not dbpath:
        dbpath = get_package_dbpath()
    return create_engine("sqlite:///{path:s}".format(path=dbpath), echo=False)


def get_session(dbpath=None):
    """Return the database session connection."""

    engine = get_engine(dbpath=dbpath)
    db_session = sessionmaker(bind=engine, autoflush=False, autocommit=False)
    return db_session()



"""
Electronegativity scale formulas
"""

import math
from typing import List

import numpy as np

from .utils import n_effective


# Rydberg constant
RY = 13.605693009


def allred_rochow(zeff: float, radius: float) -> float:
    """
    Calculate the electronegativity of an atom according to the definition
    of Allred and Rochow

    Args:
        zeff: effective nuclear charge
        radius: value of the radius
    """

    return zeff / math.pow(radius, 2)


def cottrell_sutton(zeff: float, radius: float) -> float:
    """
    Calculate the electronegativity of an atom according to the definition
    of Allred and Rochow

    Args:
        zeff: effective nuclear charge
        radius: value of the radius
    """

    return math.sqrt(zeff / radius)


def gordy(zeff: float, radius: float) -> float:
    """
    Calculate the electronegativity of an atom according to the definition
    of Allred and Rochow

    Args:
        zeff: effective nuclear charge
        radius: value of the radius
    """

    return zeff / radius


def li_xue(ionization_energy: float, radius: float, valence_pqn: int) -> float:
    """
    Calculate the electronegativity of an atom according to the definition
    of Li and Xue

    Args:
        charge: Charge of the ion
        radius: Type of radius to be used in the calculation, either `crystal_radius` as recommended in the paper or `ionic_radius`
        valence_pqn: valence principal quantum number
    """

    return (
        n_effective(valence_pqn, source="zhang")
        * math.sqrt(ionization_energy / RY)
        * 100.0
        / radius
    )


def martynov_batsanov(ionization_energies: List[float]) -> float:
    """
    Calculates the electronegativity value according to Martynov and
    Batsanov as the average of the ionization energies of the valence
    electrons

    Args:
        ionization_energies: ionization energies for the valence electrons


    .. math::

       \chi_{MB} = \sqrt{\\frac{1}{n_{v}}\sum^{n_{v}}_{k=1} I_{k}}

    where:

    - :math:`n_{v}` is the number of valence electrons and
    - :math:`I_{k}` is the :math:`k` th ionization potential.
    """

    return np.sqrt(np.array(ionization_energies).mean())


def mulliken(
    ionization_energy: float,
    electron_affinity: float,
    missing_is_zero: bool = False,
    allow_negative_ea: bool = False,
) -> float:
    """
    Return the absolute electronegativity (Mulliken scale), calculated as

    Args:
        ionization_energy: ionization energy
        electron_affinity: electron affinity
        missing_is_zero: missing values are substituted with zero
        allow_negative_ea: if `True` negative EA values will be allowed

    .. math::

       \chi = \\frac{I + A}{2}

    where:

    - :math:`I` is the ionization energy,
    - :math:`A` is the electron affinity
    """

    if ionization_energy is not None:
        if (
            electron_affinity is not None
            and electron_affinity < 0.0
            and allow_negative_ea
        ):
            return (ionization_energy + electron_affinity) * 0.5
        elif electron_affinity is not None or missing_is_zero:
            return ionization_energy * 0.5
    else:
        return None


def nagle(nvalence: int, polarizability: float) -> float:
    """
    Calculate the electronegativity of an atom according to the definition
    of Nagle

    Args:
        nvalence: number of valence electrons
        polarizability: dipole polarizability
    """

    return math.pow(nvalence / polarizability, 1.0 / 3.0)


def sanderson(radius: float, noble_gas_radius: float) -> float:
    """
    Calculate Sanderson's electronegativity

    Args:
        radius: radius value for the element
        noble_gas_radius: value of the radius of a hypothetical noble gas with
        the atomic number of element for which electronegativity is calculated

    .. math::

        \chi = \\frac{AD}{AD_{\\text{ng}}}

    """

    return math.pow(noble_gas_radius / radius, 3)


def generic(zeff: float, radius: float, rpow: float = 1, apow: float = 1) -> float:
    """
    Calculate the electronegativity from a general formula

    Args:
        zeff: effective nuclear charge
        radius: radius value for the element
        rpow: power to raise the radius to (see equation below)
        apow: power to raise the fraction to (see equation below)

    .. math::

        \chi = \left(\\frac{Z_{\\text{eff}}}{r^{\\beta}}\\right)^{\\alpha}

    where:

    - :math:`Z_{\\text{eff}}` is the effective nuclear charge
    - :math:`r` is the covalent radius
    - :math:`\\alpha,\\beta` parameters
    """

    return math.pow(zeff / math.pow(radius, rpow), apow)



# -*- coding: utf-8 -*-

from __future__ import print_function

import argparse
import textwrap
import colorama
from pyfiglet import Figlet
from mendeleev import element
from mendeleev.fetch import fetch_table


def clielement():
    """
    CLI for convenient printing of properties for a given element
    """

    colorama.init(autoreset=True)

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "element", help="Element identifier: symbol, name or atomic number"
    )
    args = parser.parse_args()

    try:
        args.element = int(args.element)
    except ValueError:
        pass

    e = element(args.element)

    f = Figlet("dotmatrix", justify="center")
    symbol = f.renderText(e.symbol)

    table = fetch_table("elements")
    et = table[table["symbol"] == e.symbol].transpose()
    et.drop("description", inplace=True)
    et.drop("sources", inplace=True)
    et.drop("uses", inplace=True)

    et.index = et.index.str.replace("_", " ").str.capitalize()
    et.sort_index(inplace=True)

    # print the data

    print(colorama.Fore.RED + symbol)

    if e.description is not None:
        print(colorama.Fore.BLUE + "Description\n===========\n")
        print("\n".join("  " + s for s in textwrap.wrap(e.description, 70)))

    if e.sources is not None:
        print(colorama.Fore.BLUE + "\nSources\n=======\n")
        print("\n".join("  " + s for s in textwrap.wrap(e.sources, 70)))

    if e.uses is not None:
        print(colorama.Fore.BLUE + "\nUses\n====\n")
        print("\n".join("  " + s for s in textwrap.wrap(e.uses, 70)))

    print(colorama.Fore.GREEN + "\nProperties\n==========\n")
    print(et.to_string(justify="left", header=False))


if __name__ == "__main__":

    clielement()



from statistics import mean
from mendeleev import element


class Ion:
    """
    Class representating atomic ions
    """

    __element_attributes__ = [
        "atomic_number",
        "block",
        "group",
        "series",
        "period",
        "mass",
        "symbol",
    ]

    def __init__(self, label, q=1):

        self._element = element(label)
        self.q = q

    @property
    def q(self):
        return self._q

    @q.setter
    def q(self, value):
        if value == 0:
            raise ValueError("expecting change other than 0, got {}".format(value))
        elif value > self.Z:
            raise ValueError(
                "ionic charge ({}) cannot be larger than atomic number ({})".format(
                    value, self.Z
                )
            )
        else:
            self._q = int(value)

    @property
    def Z(self):
        return self._element.atomic_number

    @property
    def charge(self):
        return self._q

    @property
    def electrons(self):
        return self.Z - self.q

    @property
    def name(self):
        sign = "+" if self.charge > 0 else "-"
        return "{} {}{} ion".format(self._element.name, self.charge, sign)

    @property
    def ie(self):
        return self._element.ionenergies[self.q + 1]

    @property
    def ea(self):
        return self._element.ionenergies[self.q]

    @property
    def radius(self):
        return [r for r in self._element.ionic_radii if r.charge == self.charge]

    def unicode_ion_symbol(self) -> str:
        """
        Return a unicode string symbol of the ion
        """

        superscripts = {
            "+": u"\u207A",
            "-": u"\u207B",
            "0": u"\u2070",
            "1": u"\u00B9",
            "2": u"\u00B2",
            "3": u"\u00B3",
            "4": u"\u2074",
            "5": u"\u2075",
            "6": u"\u2076",
            "7": u"\u2077",
            "8": u"\u2078",
            "9": u"\u2079",
        }
        table = str.maketrans(superscripts)
        template = "+" if self.charge > 0 else "-"

        if abs(self.charge) != 1:
            template = str(abs(self.charge)) + template

        return self.symbol + template.translate(table)

    def ionic_potential(self, radius_most_reliable: bool = True) -> float:
        """
        Calculate the ionic potential

        Args:
            radius_most_reliable : flag to use the most reliable ionic radius,
                default is `True`

        """

        if radius_most_reliable:
            radius = mean([r.ionic_radius for r in self.radius if r.most_reliable])
        else:
            radius = mean([r.ionic_radius for r in self.radius])
        return self.q / radius

    def __getattr__(self, name):

        if name in Ion.__element_attributes__:
            return getattr(self._element, name)
        else:
            raise AttributeError(
                "'{}' is not an attribute of '{}'".format(name, self.__class__.__name__)
            )

    def __repr__(self):
        return self.unicode_ion_symbol()



# -*- coding: utf-8 -*-

"""module specifying the database models"""

from typing import Any, Dict, List, Tuple
from operator import attrgetter

import numpy as np
from sqlalchemy import Column, Boolean, Integer, String, Float, ForeignKey
from sqlalchemy.orm import relationship, reconstructor
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.hybrid import hybrid_property, hybrid_method

from .electronegativity import (
    allred_rochow,
    cottrell_sutton,
    gordy,
    nagle,
    li_xue,
    martynov_batsanov,
    mulliken,
    sanderson,
)
from .db import get_session
from .econf import ElectronicConfiguration, get_l, ORBITALS
from .utils import coeffs


__all__ = [
    "Element",
    "IonizationEnergy",
    "IonicRadius",
    "OxidationState",
    "Isotope",
    "Series",
    "ScreeningConstant",
]


Base = declarative_base()


class Element(Base):
    """
    Chemical element.

    For full list of available data with references see :doc:`data`.

    Args:
        abundance_crust (float): Abundance in the earth's crust in mg/kg
        abundance_sea (float): Abundance in the seas in mg/L
        annotation (str): Annotations regarding the data
        atomic_number (int): Atomic number
        atomic_radius (float): Atomic radius in pm
        atomic_radius_rahm (float): Atomic radius by Rahm et al. in pm
        atomic_volume (float): Atomic volume in cm3/mol
        atomic_weight (float): Relative atomic weight as the ratio of the average mass of atoms
            of the element to 1/12 of the mass of an atom of 12C
        block (str): Block in periodic table, s, p, d, f
        boiling_point (float): Boiling temperature in K
        c6 (float): C_6 dispersion coefficient in a.u. from X. Chu & A. Dalgarno, J. Chem. Phys.,
            121(9), 4083–4088 (2004) doi:10.1063/1.1779576, and the value for
            Hydrogen was taken from K. T. Tang, J. M. Norbeck and P. R. Certain,
            J. Chem. Phys. 64, 3063 (1976), doi:10.1063/1.432569
        c6_gb (float): C_6 dispersion coefficient in a.u. from Gould, T., & Bučko, T. (2016).
            JCTC, 12(8), 3603–3613. http://doi.org/10.1021/acs.jctc.6b00361
        cas (str): Chemical Abstracts Service identifier
        covalent_radius_bragg (float): Covalent radius in pm from
        covalent_radius_cordero (float): Covalent radius in pm from Cordero, B., Gómez, V., Platero-Prats, A.
            E., Revés, M., Echeverría, J., Cremades, E., … Alvarez, S. (2008).
            Covalent radii revisited. Dalton Transactions, (21), 2832.
            doi:10.1039/b801115j
        covalent_radius_pyykko (float): Single bond covalent radius in pm Pyykkö, P., & Atsumi, M. (2009).
            Molecular Single-Bond Covalent Radii for Elements 1-118.
            Chemistry - A European Journal, 15(1), 186–197.
            doi:10.1002/chem.200800987
        covalent_radius_pyykko_double (float): Double bond covalent radius in pm from P. Pyykkö et al.
        covalent_radius_pyykko_triple (float): Triple bond covalent radius in pm from P. Pyykkö et al.
        cpk_color (str): CPK color of the atom in HEX, see http://jmol.sourceforge.net/jscolors/#color_U
        density (float): Density at 295K in g/cm3
        description (str): Short description of the element
        dipole_polarizability (float): Dipole polarizability in atomic units
        dipole_polarizability_unc (float): Uncertainty of the dipole polarizability
        discoverers (str): The discoverers of the element
        discovery_location (str): The location where the element was discovered
        discovery_year (int): The year the element was discovered
        electron_affinity (float): Electron affinity in eV
        electrophilicity (float): Parr's electrophilicity index
        econf (str): Ground state electron configuration
        evaporation_heat (float): Evaporation heat in kJ/mol
        fusion_heat (float): Fusion heat in kJ/mol
        gas_basicity (float): Gas basicity
        geochemical_class (str): Geochemical classification of the elements
        glawe_number (int): Glawe number (scale)
        goldschmidt_class (str): Goldschmidt classification of the elements
        group_id (int): Group number
        heat_of_formation (float): Heat of formation in kJ/mol
        is_monoisotopic (bool): A flag marking if the element is monoisotopic
        jmol_color (str): Color of the atom as used in Jmol, in HEX,
            see http://jmol.sourceforge.net/jscolors/#color_U
        lattice_constant (float): Lattice constant in ang
        lattice_structure (str): Lattice structure code
        mass (float): Relative atomic mass. Ratio of the average mass of atoms
            of the element to 1/12 of the mass of an atom of 12C
        mendeleev_number (int): Mendeleev number
        melting_point (float): Melting temperature in K
        metallic_radius (float): Single-bond metallic radius or metallic radius, have been
            calculated by Pauling using interatomic distances and an
            equation relating such distances with bond number
        metallic_radius_c12 (float): Metallic radius obtained by Pauling with an assumed number of
            nearest neighbors equal to 12
        molcas_gv_color (str): Color of an atom in HEX from MOLCAS GV http://www.molcas.org/GV/
        name (str): Name in English
        name_origin (str): Origin of the name
        period (int): Period in periodic table
        pettifor_number (int): Pettifor scale
        proton_affinity (Float): Proton affinity
        series (int): Index to chemical series
        sources (str): Sources of the element
        specific_heat (float): Specific heat in J/g mol @ 20 C
        symbol (str): Chemical symbol
        thermal_conductivity (float): Thermal conductivity in @/m K @25 C
        uses (str): Uses of the element
        vdw_radius (float): Van der Waals radius in pm from W. M. Haynes, Handbook of Chemistry and
            Physics 95th Edition, CRC Press, New York, 2014, ISBN-10: 1482208679,
            ISBN-13: 978-1482208672.
        vdw_radius_bondi (float): Van der Waals radius according to Bondi in pm
        vdw_radius_truhlar (float): Van der Waals radius according to Truhlar in pm
        vdw_radius_rt (float): Van der Waals radius according to Rowland and Taylor in pm
        vdw_radius_batsanov (float): Van der Waals radius according to Batsanov in pm
        vdw_radius_dreiding (float): Van der Waals radius from the DREIDING force field in pm
        vdw_radius_uff (float): Van der Waals radius from the UFF in pm
        vdw_radius_mm3 (float): Van der Waals radius from MM3 in pm
        oxistates (list): Oxidation states
        ionenergies (dict): Ionization energies in eV parsed from
            http://physics.nist.gov/cgi-bin/ASD/ie.pl on April 13, 2015
    """

    __tablename__ = "elements"

    abundance_crust = Column(Float)
    abundance_sea = Column(Float)
    annotation = Column(String)
    atomic_number = Column(Integer, primary_key=True)
    atomic_radius = Column(Float)
    atomic_radius_rahm = Column(Float)
    atomic_volume = Column(Float)
    atomic_weight = Column(Float)
    atomic_weight_uncertainty = Column(Float)
    block = Column(String)
    boiling_point = Column(Float)
    cas = Column(String)
    covalent_radius_bragg = Column(Float)
    covalent_radius_cordero = Column(Float)
    covalent_radius_pyykko = Column(Float)
    covalent_radius_pyykko_double = Column(Float)
    covalent_radius_pyykko_triple = Column(Float)
    c6 = Column(Float)
    c6_gb = Column(Float)
    cpk_color = Column(String)
    density = Column(Float)
    description = Column(String)
    dipole_polarizability = Column(Float)
    dipole_polarizability_unc = Column(Float)
    discoverers = Column(String)
    discovery_location = Column(String)
    discovery_year = Column(Integer)
    electron_affinity = Column(Float)
    en_allen = Column(Float)
    en_ghosh = Column(Float)
    en_pauling = Column(Float)
    econf = Column("electronic_configuration", String)
    evaporation_heat = Column(Float)
    fusion_heat = Column(Float)
    gas_basicity = Column(Float)
    geochemical_class = Column(String)
    glawe_number = Column(Integer)
    goldschmidt_class = Column(String)
    group_id = Column(Integer, ForeignKey("groups.group_id"))
    group = relationship("Group", uselist=False, lazy="subquery")
    heat_of_formation = Column(Float)
    is_monoisotopic = Column(Boolean)
    is_radioactive = Column(Boolean)
    jmol_color = Column(String)
    lattice_constant = Column(Float)
    lattice_structure = Column(String)
    melting_point = Column(Float)
    mendeleev_number = Column(Integer)
    metallic_radius = Column(Float)
    metallic_radius_c12 = Column(Float)
    molcas_gv_color = Column(String)
    name = Column(String)
    name_origin = Column(String)
    period = Column(Integer)
    pettifor_number = Column(Integer)
    proton_affinity = Column(Float)
    _series_id = Column("series_id", Integer, ForeignKey("series.id"))
    _series = relationship("Series", uselist=False, lazy="subquery")
    series = association_proxy("_series", "name")
    sources = Column(String)
    specific_heat = Column(Float)
    symbol = Column(String)
    thermal_conductivity = Column(Float)
    uses = Column(String)
    vdw_radius = Column(Float)
    vdw_radius_alvarez = Column(Float)
    vdw_radius_bondi = Column(Float)
    vdw_radius_truhlar = Column(Float)
    vdw_radius_rt = Column(Float)
    vdw_radius_batsanov = Column(Float)
    vdw_radius_dreiding = Column(Float)
    vdw_radius_uff = Column(Float)
    vdw_radius_mm3 = Column(Float)

    ionic_radii = relationship("IonicRadius", lazy="subquery")
    _ionization_energies = relationship("IonizationEnergy", lazy="subquery")
    _oxidation_states = relationship("OxidationState", lazy="subquery")
    isotopes = relationship("Isotope", lazy="subquery")
    screening_constants = relationship("ScreeningConstant", lazy="subquery")

    @reconstructor
    def init_on_load(self):
        "Initialize the ElectronicConfiguration class as attribute of self"

        self.ec = ElectronicConfiguration(self.econf)

    @hybrid_property
    def ionenergies(self) -> Dict[int, float]:
        """
        Return a dict with ionization degree as keys and ionization energies
        in eV as values.
        """

        return {ie.degree: ie.energy for ie in self._ionization_energies}

    @hybrid_property
    def oxistates(self) -> List[int]:
        """Return the oxidation states as a list of ints"""

        return [os.oxidation_state for os in self._oxidation_states]

    @hybrid_property
    def sconst(self) -> Dict[Tuple[int], float]:
        """
        Return a dict with screening constants with tuples (n, s) as keys and
        screening constants as values"""

        return {(x.n, x.s): x.screening for x in self.screening_constants}

    @hybrid_property
    def electrons(self) -> int:
        """Return the number of electrons."""

        return self.atomic_number

    @hybrid_property
    def neutrons(self) -> int:
        """
        Return the number of neutrons of the most abundant natural stable
        isotope.
        """

        return self.mass_number - self.protons

    @hybrid_property
    def protons(self) -> int:
        """Return the number of protons."""

        return self.atomic_number

    @hybrid_property
    def mass(self) -> float:
        """
        Return the `atomic_weight` if defined or mass number otherwise.
        """

        return self.atomic_weight

    @hybrid_property
    def mass_number(self) -> int:
        """
        Return the mass number of the most abundant natural stable isotope
        """

        if len(self.isotopes) <= 0:
            return int(self.atomic_weight)

        lwithabu = [i for i in self.isotopes if i.abundance is not None]
        if lwithabu:
            return max(lwithabu, key=attrgetter("abundance")).mass_number
        else:
            return self.isotopes[0].mass_number

    def mass_str(self) -> str:
        """String representation of atomic weight"""

        if self.atomic_weight_uncertainty is None:
            if self.is_radioactive:
                return "[{aw:.0f}]".format(aw=self.atomic_weight)
            else:
                return "{aw:.3f}".format(aw=self.atomic_weight)
        else:
            dec = np.abs(
                np.floor(np.log10(np.abs(self.atomic_weight_uncertainty)))
            ).astype(int)
            dec = min(dec, 5)
            if self.is_radioactive:
                return "[{aw:.{dec}f}]".format(aw=self.atomic_weight, dec=dec)
            else:
                return "{aw:.{dec}f}".format(aw=self.atomic_weight, dec=dec)

    @hybrid_property
    def covalent_radius(self) -> float:
        """
        Return the default covalent radius which is covalent_radius_pyykko
        """

        return self.covalent_radius_pyykko

    @hybrid_method
    def hardness(self, charge: int = 0) -> float:
        """
        Return the absolute hardness, calculated as

        Args:
          charge:  Charge of the cation for which the hardness will be calculated

        .. math::

           \eta = \\frac{IE - EA}{2}

        where:

        - :math:`IE` is the ionization energy,
        - :math`EA` is the electron affinity

        """

        if charge == 0:
            if (
                self.ionenergies.get(1, None) is not None
                and self.electron_affinity is not None
            ):
                return (self.ionenergies[1] - self.electron_affinity) * 0.5
            else:
                return None
        elif charge > 0:
            if (
                self.ionenergies.get(charge + 1, None) is not None
                and self.ionenergies.get(charge, None) is not None
            ):
                return (self.ionenergies[charge + 1] - self.ionenergies[charge]) * 0.5
            else:
                return None
        elif charge < 0:
            raise ValueError(
                "Charge has to be a non-negative integer, got: {}".format(charge)
            )

    @hybrid_method
    def softness(self, charge: int = 0) -> float:
        """
        Return the absolute softness, calculated as

        Args:
          charge: Charge of the cation for which the hardness will be calculated

        .. math::

           S = \\frac{1}{2\eta}

        where :

        - :math:`\eta` is the absolute hardness

        """

        eta = self.hardness(charge=charge)

        if eta is None:
            return None
        else:
            return 1.0 / (2.0 * eta)

    def zeff(
        self, n: int = None, o: str = None, method: str = "slater", alle: bool = False
    ) -> float:
        """
        Return the effective nuclear charge for ``(n, s)``

        Args:
            method: Method to calculate the screening constant, the choices are

                - `slater`, for Slater's method as in Slater, J. C. (1930).
                    Atomic Shielding Constants. Physical Review, 36(1), 57–64.
                    `doi:10.1103/PhysRev.36.57 <http://www.dx.doi.org/10.1103/PhysRev.36.57>`_
                - `clementi` for values of screening constants from Clementi, E.,
                    & Raimondi, D. L. (1963). Atomic Screening Constants from SCF
                    Functions. The Journal of Chemical Physics, 38(11), 2686.
                    `doi:10.1063/1.1733573 <http://www.dx.doi.org/10.1063/1.1733573>`_
                    and Clementi, E. (1967). Atomic Screening Constants from SCF
                    Functions. II. Atoms with 37 to 86 Electrons. The Journal of
                    Chemical Physics, 47(4), 1300.
                    `doi:10.1063/1.1712084 <http://www.dx.doi.org/10.1063/1.1712084>`_

            n: Principal quantum number

            o: Orbital label, (s, p, d, ...)

            alle: Use all the valence electrons, i.e. calculate screening for an
                extra electron when method='slater', if method='clementi' this
                option is ignored
        """

        # identify the valence s,p vs d,f
        if n is None:
            n = self.ec.max_n()
        elif not isinstance(n, int):
            raise ValueError("<n> should be an integer, got: {}".format(type(n)))

        if o is None:
            # take the shell with max `l` for a given `n`
            o = ORBITALS[max(get_l(x[1]) for x in self.ec.conf.keys() if x[0] == n)]
        elif o not in ORBITALS:
            raise ValueError("<s> should be one of {}".format(", ".join(ORBITALS)))

        if method.lower() == "slater":
            return self.atomic_number - self.ec.slater_screening(n=n, o=o, alle=alle)
        elif method.lower() == "clementi":
            sc = self.sconst.get((n, o), None)
            if sc is not None:
                return self.atomic_number - self.sconst.get((n, o), None)
            else:
                return sc
        else:
            raise ValueError("<method> should be one of {}".format("slater, clementi"))

    def electrophilicity(self) -> float:
        """
        Calculate electrophilicity index

        .. math::

           \\omega = \\frac{\\mu}{2\eta}
        """

        ip = self.ionenergies.get(1, None)
        ea = self.electron_affinity

        if ip is not None and ea is not None:
            return (ip + ea) ** 2 / (8.0 * (ip - ea))
        else:
            return None

    def electronegativity_scales(self, name: str = None) -> List[str]:
        "Available electronegativity scales"

        scales = {
            "allen": self.electronegativity_allen,
            "allred-rochow": self.electronegativity_allred_rochow,
            "cottrell-sutton": self.electronegativity_cottrell_sutton,
            "ghosh": self.electronegativity_ghosh,
            "gordy": self.electronegativity_gordy,
            "li-xue": self.electronegativity_li_xue,
            "martynov-batsanov": self.electronegativity_martynov_batsanov,
            "mulliken": self.electronegativity_mulliken,
            "nagle": self.electronegativity_nagle,
            "pauling": self.electronegativity_pauling,
            "sanderson": self.electronegativity_sanderson,
        }

        if name:
            if name in scales:
                return scales[name]
            else:
                raise ValueError(
                    f"scale: '{name}' not found, available scales are: {', '.join(scales.keys())}"
                )

        return list(sorted(scales.keys()))

    def electronegativity(self, scale: str = "pauling", **kwargs) -> float:
        """
        Calculate the electronegativity using one of the methods

        Args:
            scale : Name of the electronegativity scale, one of
            kwargs: keyword arguments that are passed to compute a specific electronegativity

        """

        return self.electronegativity_scales(name=scale)(**kwargs)

    def electronegativity_allen(self) -> float:
        "Allen's electronegativity"
        return self.en_allen

    def electronegativity_allred_rochow(self, radius="covalent_radius_pyykko") -> float:
        "Allred-Rochow's electronegativity"
        return allred_rochow(self.zeff(), getattr(self, radius))

    def electronegativity_cottrell_sutton(
        self, radius="covalent_radius_pyykko"
    ) -> float:
        "Cottrell-Sutton's electronegativity"
        return cottrell_sutton(self.zeff(), getattr(self, radius))

    def electronegativity_gordy(self, radius="covalent_radius_pyykko") -> float:
        "Gordy's electronegativity"
        return gordy(self.zeff(), getattr(self, radius))

    def electronegativity_ghosh(self) -> float:
        "Ghosh's electronegativity"
        return self.en_ghosh

    def electronegativity_li_xue(
        self, charge: int = 1, radius: str = "crystal_radius"
    ) -> float:
        """
        Calculate the electronegativity of an atom according to the definition
        of Li and Xue

        Args:
            charge : charge of the ion
            radius : type of radius to be used in the calculation, either
                `crystal_radius` as recommended in the paper or `ionic_radius`

        Returns:
            out (dict): dictionary with electronegativities as values and
                coordination string as keys or tuple of coordination and spin
                if the ion is LS or HS
        """

        if (not isinstance(charge, int)) or (charge == 0):
            raise ValueError(
                "charge should be a nonzero initeger, got: {}".format(charge)
            )

        if radius not in ["ionic_radius", "crystal_radius"]:
            raise ValueError(
                f"radius: '{radius}' not found, available values are: 'ionic_radius', 'crystal_radius'"
            )

        ie = self.ionenergies.get(charge, None)

        radii = [
            (ir.coordination, ir.spin, getattr(ir, radius))
            for ir in self.ionic_radii
            if ir.charge == charge
        ]

        return {
            (coordination_number, spin): li_xue(ie, crystal_radius, self.ec.max_n())
            for coordination_number, spin, crystal_radius in radii
        }

    def electronegativity_martynov_batsanov(self) -> float:
        """
        Calculates the electronegativity value according to Martynov and
        Batsanov as the average of the ionization energies of the valence
        electrons

        .. math::

           \chi_{MB} = \sqrt{\\frac{1}{n_{v}}\sum^{n_{v}}_{k=1} I_{k}}

        where:

        - :math:`n_{v}` is the number of valence electrons
        - :math:`I_{k}` is the :math:`k` th ionization potential.

        """

        ionenergies = [
            self.ionenergies.get(i, None)
            for i in range(1, self.nvalence(method="simple") + 1)
        ]

        if all(ionenergies):
            return martynov_batsanov(ionenergies)
        else:
            return None

    def electronegativity_mulliken(
        self,
        charge: int = 0,
        missing_is_zero: bool = False,
        allow_negative_ea: bool = False,
    ) -> float:
        """
        Return the absolute electronegativity (Mulliken scale), calculated as

        Args:
            charge: charge of the ion
            missing_is_zero: missing values are substituted with zero
            allow_negative_ea: if `True` negative EA values will be allowed

        .. math::

           \chi = \\frac{I + A}{2}

        where :math:`I` is the ionization energy and :math:`A` is the electron
        affinity
        """

        if charge == 0:
            ip = self.ionenergies.get(1, None)
            ea = self.electron_affinity
        elif charge > 0:
            ip = self.ionenergies.get(charge + 1, None)
            ea = self.ionenergies.get(charge, None)
        else:
            raise ValueError(
                "Charge has to be a non-negative integer, got: {}".format(charge)
            )
        return mulliken(
            ip, ea, missing_is_zero=missing_is_zero, allow_negative_ea=allow_negative_ea
        )

    def electronegativity_nagle(self) -> float:
        "Nagle's electronegativity"

        if self.dipole_polarizability is not None:
            return nagle(self.nvalence(), self.dipole_polarizability)

    def electronegativity_pauling(self) -> float:
        "Pauling's electronegativity"
        return self.en_pauling

    def electronegativity_sanderson(self, radius="covalent_radius_pyykko") -> float:
        """
        Sanderson electronegativity

        Args:
            radius : radius to use in the calculation
        """
        # estimate the radius of a corresponding noble gas
        noble_gas_radius = estimate_from_group(self.atomic_number, radius)
        return sanderson(getattr(self, radius), noble_gas_radius)

    def nvalence(self, method=None) -> int:
        """
        Return the number of valence electrons
        """

        return self.ec.nvalence(self.block, method=method)

    def oxides(self) -> List[str]:
        """
        Return a lsit of possible oxides based on the oxidation number
        """

        oxide_coeffs = [coeffs(ox) for ox in self.oxistates if ox > 0]
        # convert to strings and replace 1 with empty string
        normal_coeffs = [[str(c) if c != 1 else "" for c in t] for t in oxide_coeffs]
        return [f"{self.symbol}{cme}O{co}" for cme, co in normal_coeffs]

    def __str__(self):
        return "{0} {1} {2}".format(self.atomic_number, self.symbol, self.name)

    def __repr__(self):
        return "%s(\n%s)" % (
            self.__class__.__name__,
            " ".join(
                "\t%s=%r,\n" % (key, getattr(self, key))
                for key in sorted(self.__dict__.keys())
                if not key.startswith("_")
            ),
        )


def fetch_attrs_for_group(attrs: List[str], group: int = 18) -> Tuple[List[Any]]:
    """
    A convenience function for getting a specified attribute for all
    the memebers of a given group.

    Args:
        attr : Attribute of `Element` to retrieve for all group members

    Returns:
        data (dict): Dictionary with noble gas atomic numbers as keys and values of the
            `attr` as values
    """

    session = get_session()
    members = (
        session.query(Element)
        .filter(Element.group_id == group)
        .order_by(Element.atomic_number)
        .all()
    )

    results = tuple([getattr(member, attr) for member in members] for attr in attrs)
    session.close()
    return results


def estimate_from_group(
    atomic_number, attr_name, group: int = 18, deg: int = 1
) -> float:
    """
    Evaluate a value `attribute` for element by interpolation or
    extrapolation of the data points from elements from `group`.

    Args:
        atomic_number: value for which the property will be evaluated
        attr_name: attribute to be estimated
        group: periodic table group number
        deg: degree of the polynomial used in the extrapolation beyond
            the provided data points
    """

    xref, yref = fetch_attrs_for_group(["atomic_number", attr_name], group=group)

    x = atomic_number
    xref = np.array(xref)
    yref = np.array(yref)
    if xref.min() <= x <= xref.max():
        return np.interp([x], xref, yref)

    if x < xref.min():
        xslice = xref[:3]
        yslice = yref[:3]
    elif x > xref.max():
        xslice = xref[-3:]
        yslice = yref[-3:]

    fit = np.polyfit(xslice, yslice, deg)
    fn = np.poly1d(fit)
    return fn(x)


class IonicRadius(Base):
    """
    Effective ionic radii and crystal radii in pm retrieved from [1]_.

    .. [1] Shannon, R. D. (1976). Revised effective ionic radii and systematic
       studies of interatomic distances in halides and chalcogenides. Acta
       Crystallographica Section A.
       `doi:10.1107/S0567739476001551 <http://www.dx.doi.org/10.1107/S0567739476001551>`_

    Args:
        atomic_number (int): Atomic number
        charge (int):  Charge of the ion
        econf (str): Electronic configuration of the ion
        coordination (str): Type of coordination
        spin (str): Spin state: HS - high spin, LS - low spin
        crystal_radius (float): Crystal radius in pm
        ionic_radius (float): Ionic radius in pm
        origin (str): Source of the data
        most_reliable (bool): Most reliable value (see reference)
    """

    __tablename__ = "ionicradii"

    id = Column(Integer, primary_key=True)
    atomic_number = Column(Integer, ForeignKey("elements.atomic_number"))
    charge = Column(Integer)
    econf = Column(String)
    coordination = Column(String)
    spin = Column(String)
    crystal_radius = Column(Float)
    ionic_radius = Column(Float)
    origin = Column(String)
    most_reliable = Column(Boolean)

    def __str__(self):
        out = ["{0}={1:>4d}", "{0}={1:5s}", "{0}={1:>6.3f}", "{0}={1:>6.3f}"]
        keys = ["charge", "coordination", "crystal_radius", "ionic_radius"]
        return ", ".join(o.format(k, getattr(self, k)) for o, k in zip(out, keys))

    def __repr__(self):
        return "%s(\n%s)" % (
            self.__class__.__name__,
            " ".join(
                "\t%s=%r,\n" % (key, getattr(self, key))
                for key in sorted(self.__dict__.keys())
                if not key.startswith("_")
            ),
        )


class IonizationEnergy(Base):
    """
    Ionization energy of an element

    Args:
        atomic_number (int): Atomic number
        degree (int): Degree of ionization with respect to neutral atom
        energy (float): Ionization energy in eV parsed from
            http://physics.nist.gov/cgi-bin/ASD/ie.pl on April 13, 2015
    """

    __tablename__ = "ionizationenergies"

    id = Column(Integer, primary_key=True)
    atomic_number = Column(Integer, ForeignKey("elements.atomic_number"))
    degree = Column(Integer)
    energy = Column(Float)

    def __str__(self):

        return "{0:5d} {1:10.5f}".format(self.degree, self.energy)

    def __repr__(self):

        return "<IonizationEnergy(atomic_number={a:5d}, degree={d:3d}, energy={e:10.5f})>".format(
            a=self.atomic_number, d=self.degree, e=self.energy
        )


class OxidationState(Base):
    """
    Oxidation states of an element

    Args:
      atomic_number (int): Atomic number
      oxidation_state (int): Oxidation state
    """

    __tablename__ = "oxidationstates"

    id = Column(Integer, primary_key=True)
    atomic_number = Column(Integer, ForeignKey("elements.atomic_number"))
    oxidation_state = Column(Integer)

    def __repr__(self):

        return "<OxidationState(atomic_number={a:5d}, oxidation_state={o:5d})>".format(
            a=self.atomic_number, o=self.oxidation_state
        )


class Group(Base):
    """
    Name of the group in the periodic table.

    Args:
        group_id (int): group number
        symbol: (str): group symbol
        name (str): group name
    """

    __tablename__ = "groups"

    group_id = Column(Integer, primary_key=True)
    symbol = Column(String)
    name = Column(String)

    def __repr__(self):

        return "<Group(symbol={s:s}, name={n:s})>".format(s=self.symbol, n=self.name)


class Series(Base):
    """
    Name of the series in the periodic table.

    Args:
        name (str): Name of the series
        color (str): The HEX representation of a color of the series, the colors were
            obtained from `ColorBrewer <http://colorbrewer2.org/?type=qualitative&scheme=Paired&n=10>`_
            the qualitative 10-class paired colormap
    """

    __tablename__ = "series"

    id = Column(Integer, primary_key=True)
    name = Column(String)
    color = Column(String)

    def __repr__(self):

        return "<Series(name={n:s}, color={c:s})>".format(n=self.name, c=self.color)


class Isotope(Base):
    """
    Isotope

    Args:
        abundance (float): Abundance of the isotope
        atomic_number (int): Atomic number
        half_life (float): Half life time
        half_life_unit (str): Unit for the half life time
        is_radioactive (bool): A flag marking wheather the isotope is radioactive
        mass (float): Mass of the isotope
        mass_number (int): Mass number of the isotope
        mass_uncertainty (float): Uncertainty of the mass value
    """

    __tablename__ = "isotopes"

    id = Column(Integer, primary_key=True)
    abundance = Column(Float)
    atomic_number = Column(Integer, ForeignKey("elements.atomic_number"))
    g_factor = Column(Float)
    half_life = Column(Float)
    half_life_unit = Column(String)
    is_radioactive = Column(Boolean)
    mass = Column(Float)
    mass_number = Column(Integer)
    mass_uncertainty = Column(Float)
    spin = Column(Float)
    quadrupole_moment = Column(Float)

    def __str__(self):

        afmt = "5.3f"
        mfmt = "10.5f"

        if self.mass is None:
            mfmt = ""

        if self.abundance is None:
            afmt = ""

        return "{0:5d} {1:5d} {2:{mfmt}} {3:{afmt}}".format(
            self.atomic_number,
            self.mass_number,
            self.mass,
            self.abundance,
            mfmt=mfmt,
            afmt=afmt,
        )

    def __repr__(self):

        return "<Isotope(Z={}, A={}, mass={}, abundance={})>".format(
            self.atomic_number, self.mass_number, self.mass, self.abundance
        )


class ScreeningConstant(Base):
    """
    Nuclear screening constants from Clementi, E., & Raimondi, D. L. (1963).
    Atomic Screening Constants from SCF Functions. The Journal of Chemical
    Physics, 38(11), 2686.  `doi:10.1063/1.1733573
    <http://www.dx.doi.org/10.1063/1.1733573>`_ and Clementi, E. (1967). Atomic
    Screening Constants from SCF Functions. II. Atoms with 37 to 86 Electrons.
    The Journal of Chemical Physics, 47(4), 1300.  `doi:10.1063/1.1712084
    <http://www.dx.doi.org/10.1063/1.1712084>`_

    Args::
        atomic_number (int): Atomic number
        n (int): Principal quantum number
        s (str): Subshell label, (s, p, d, ...)
        screening (float): Screening constant
    """

    __tablename__ = "screeningconstants"

    id = Column(Integer, primary_key=True)
    atomic_number = Column(Integer, ForeignKey("elements.atomic_number"))
    n = Column(Integer)
    s = Column(String)
    screening = Column(Float)

    def __str__(self):

        return "{0:4d} {1:3d} {2:s} {3:10.4f}".format(
            self.atomic_number, self.n, self.s, self.screening
        )

    def __repr__(self):

        return "<ScreeningConstant(Z={0:4d}, n={1:3d}, s={2:s}, screening={3:10.4f})>".format(
            self.atomic_number, self.n, self.s, self.screening
        )



from typing import Union, Tuple
import math


def coeffs(a: int, b: int = 2) -> Tuple[int]:
    """
    Return stoichometric coefficients from oxidation states

    Args:
        a: oxidation state of the first element
        b: oxidation state  of the second element
    """
    lcm = abs(a * b) // math.gcd(a, b)
    return lcm // a, lcm // b


def n_effective(n: int, source: str = "slater") -> Union[float, None]:
    """
    Effective principal quantum number

    Args:
        n: Principal quantum number
        source: either `slater` or `zhang`, for more information see note below.

    .. note::

       Slater's values are taken from J. A. Pople, D. L. Beveridge,
       "Approximate Molecular Orbital Theory", McGraw-Hill, 1970

       Zhang's values are taken from Zhang, Y. (1982). Electronegativities
       of elements in valence states and their applications.
       1. Electronegativities of elements in valence states.
       Inorganic Chemistry, 21(11), 3886–3889. https://doi.org/10.1021/ic00141a005
    """

    numbers = {
        "slater": {1: 1.0, 2: 2.0, 3: 3.0, 4: 3.7, 5: 4.0, 6: 4.2},
        "zhang": {1: 0.85, 2: 1.99, 3: 2.89, 4: 3.45, 5: 3.85, 6: 4.36, 7: 4.99},
    }

    if source not in numbers.keys():
        raise ValueError(
            f"source '{source}' not found, available sources are: {', '.join(numbers.keys())}"
        )

    return numbers.get(source).get(n, None)



from mendeleev.vis.bokeh import periodic_table_bokeh
from mendeleev.vis.plotly import periodic_table_plotly
from mendeleev.vis.utils import create_vis_dataframe


def periodic_table(
    attribute: str = "atomic_weight",
    height: int = 800,
    width: int = 1200,
    decimals: int = 3,
    colorby: str = "color",
    missing: str = "#ffffff",
    title: str = "Periodic Table",
    cmap: str = "RdBu_r",
    wide_layout: bool = False,
    include_f_block: bool = True,
    backend="plotly",
):

    elements = create_vis_dataframe(
        include_f_block=include_f_block, wide_layout=wide_layout
    )

    backends = {
        "plotly": periodic_table_plotly,
        "bokeh": periodic_table_bokeh,
    }

    if backend not in backends:
        raise ValueError(f"unsupported plotting backend: {backend}")

    plotter = backends[backend]
    return plotter(
        elements,
        attribute=attribute,
        height=height,
        width=width,
        decimals=decimals,
        colorby=colorby,
        missing=missing,
        title=title,
        cmap=cmap,
        wide_layout=wide_layout,
    )


from .periodictable import periodic_table
from .plotly import periodic_table_plotly
from .bokeh import periodic_table_bokeh
from .seaborn import heatmap
from .utils import create_vis_dataframe, add_tile_coordinates


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mendeleev', 'mendeleev.vis']

package_data = \
{'': ['*']}

install_requires = \
['Pygments>=2.8.0,<3.0.0',
 'SQLAlchemy>=1.3.23,<2.0.0',
 'colorama>=0.4.4,<0.5.0',
 'numpy>=1.19.5,<2.0.0',
 'pandas>=0.25.0',
 'pyfiglet>=0.8.post1,<0.9',
 'six>=1.15.0,<2.0.0']

extras_require = \
{'vis': ['plotly>=5.0,<6.0',
         'seaborn>=0.11.1,<0.12.0',
         'bokeh>=2.3.0,<3.0.0',
         'jupyter-bokeh>=3.0.2,<4.0.0']}

setup_kwargs = {
    'name': 'mendeleev',
    'version': '0.8.0',
    'description': 'Pythonic periodic table of elements',
    'long_description': '[![Documentation Status](https://readthedocs.org/projects/mendeleev/badge/)](https://mendeleev.readthedocs.org)\n[![Latest version released on PyPi](https://img.shields.io/pypi/v/mendeleev.svg?style=flat-square&label=PYPI%20version)](https://pypi.python.org/pypi/mendeleev)\n[![MIT license](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n[![image](https://anaconda.org/conda-forge/github3.py/badges/installer/conda.svg)](https://anaconda.org/lmmentel/mendeleev)\n[![Python package](https://github.com/lmmentel/mendeleev/actions/workflows/main.yml/badge.svg)](https://github.com/lmmentel/mendeleev/actions/workflows/main.yml)\n[![Total alerts](https://img.shields.io/lgtm/alerts/g/lmmentel/mendeleev.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/lmmentel/mendeleev/alerts/)\n[![Language grade: Python](https://img.shields.io/lgtm/grade/python/g/lmmentel/mendeleev.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/lmmentel/mendeleev/context:python)\n[![image](https://pepy.tech/badge/mendeleev)](https://pepy.tech/project/mendeleev)\n\n\n\n<!-- PROJECT LOGO -->\n<br />\n<p align="center">\n  <a href="https://github.com/lmmentel/mendeleev">\n    <img src="docs/source/_static/img/name_and_logo.png" alt="Logo" width="300">\n  </a>\n\n  <h3 align="center">Pythonic periodic table of elements</h3>\n\n  <p align="center">\n    A package with a convenient python API for accessing various properties\nof elements, ions and isotopes in the periodic table of elements.\n    <br />\n    <br />\n    <a href="https://github.com/lmmentel/mendeleev/issues">Report Bug</a>\n    ·\n    <a href="https://github.com/lmmentel/mendeleev/issues">Request Feature</a>\n    ·\n    <a href="https://github.com/lmmentel/mendeleev/pulls">Submit a Pull Request</a>\n  </p>\n</p>\n\n\n<!-- TABLE OF CONTENTS -->\n## Table of Contents\n\n- [Table of Contents](#table-of-contents)\n- [About the project](#about-the-project)\n  - [Interactive web app](#interactive-web-app)\n- [Installation](#installation)\n- [Documentation](#documentation)\n- [Data](#data)\n  - [Basic properties](#basic-properties)\n  - [Standardized colors schemes](#standardized-colors-schemes)\n  - [Size related properties](#size-related-properties)\n  - [Electronegativity scales](#electronegativity-scales)\n  - [Descriptive properties](#descriptive-properties)\n  - [Physical properties](#physical-properties)\n  - [Computed properties](#computed-properties)\n  - [Isotope properties](#isotope-properties)\n- [Getting started](#getting-started)\n  - [Lists of elements](#lists-of-elements)\n    - [Isotopes](#isotopes)\n  - [Accesing data tables and the database](#accesing-data-tables-and-the-database)\n  - [Command line interface utility](#command-line-interface-utility)\n- [Contributing](#contributing)\n  - [Issues](#issues)\n  - [Pull requests](#pull-requests)\n- [License](#license)\n- [Citing](#citing)\n- [Funding](#funding)\n\n---\n\nsource: [https://github.com/lmmentel/mendeleev](https://github.com/lmmentel/mendeleev)\n\n---\n\n<h2> Important! </h2>\n\nVersion 0.5.1 will be the latest version to support Python2.7.\n\n**This package has been migrated from bitbucket.**\n\n**The bitbucket repo is no longer being maintaned and development will\ncontinue at github.**\n\n---\n\n## About the project\n\nThis package provides a convenient python API for accessing various\nproperties of elements, ions and isotopes in the periodic table of\nelements.\n\nMoreover it provides an easy to use interface to\n[pandas](http://pandas.pydata.org/) and convenient visualization\nfunctionality through [bokeh](http://bokeh.pydata.org/en/latest/) that\nenables you to create customized periodic tables displaying various\nproperties.\n\n![peridic_table](docs/source/_static/img/mendeleev_periodic_series.png)\nDjango Extensions is free and always will be. It is development and maintained by developers in an Open Source manner. Any support is welcome. You could help by writing documentation, pull-requests, report issues and/or translations.\niodic trends in the periodic tables. If you want to\nlook at some examples there are a few\n[tutorials](http://mendeleev.readthedocs.io/en/stable/tutorials.html)\navailable as [jupyter notebooks](http://jupyter.org/).\n\n### Interactive web app\n\nIf you would like to explore the data available in\n[mendeleev](http://mendeleev.readthedocs.org) check out the interactive\nweb app at [mendeleev.herokuapp.com](http://mendeleev.herokuapp.com/)\nwhere you can create your own periodic tables and visualize the\nrelations between various properties of elements.\n\n![Periodic table view](docs/source/_static/img/mendeleevapp_periodic.png)\n\n![Correlations view](docs/source/_static/img/mendeleevapp_correlations.png)\n\n## Installation\n\nThe preferred installation method is with\n[conda](https://conda.io/docs/intro.html) and you can install the\npackage from the [conda-forge](https://conda-forge.org/) channel by\n\n``` {.sourceCode .bash}\nconda install -c conda-forge mendeleev=0.5.2\n```\n\nThe package can also be installed using\n[pip](https://pypi.python.org/pypi/pip)\n\n``` {.sourceCode .bash}\npip install mendeleev\n```\nor [pipenv](https://pipenv.readthedocs.io/en/latest/)\n\n``` {.sourceCode .bash}\npipenv install mendeleev\n```\n\nYou can also install the most recent version from the repository:\n\n``` {.sourceCode .bash}\npip install git+https://github.com/lmmentel/mendeleev.git\n```\n\n## Documentation\n\nFull documentation is hosted on [Read the Docs](http://mendeleev.readthedocs.org/en/latest/).\n\nThere are also tutorials available as `Jupyter <https://jupyter.org/>`_\nnotebooks on `binder <https://mybinder.org/>`_ where you can explore the examples interactively:\n\n* `Quick start <https://mybinder.org/v2/gh/lmmentel/mendeleev/master?filepath=docs%2Fsource%2Fnotebooks%2Fquick_start.ipynb>`_\n* `Bulk data access <https://mybinder.org/v2/gh/lmmentel/mendeleev/master?filepath=docs%2Fsource%2Fnotebooks%2Fbulk_data_access.ipynb>`_\n* `Electronic Configuration <https://mybinder.org/v2/gh/lmmentel/mendeleev/master?filepath=docs%2Fsource%2Fnotebooks%2Felectronic_configuration.ipynb>`_\n* `Ions <https://mybinder.org/v2/gh/lmmentel/mendeleev/master?filepath=docs%2Fsource%2Fnotebooks%2Fions.ipynb>`_\n* `Visualizations <https://mybinder.org/v2/gh/lmmentel/mendeleev/master?filepath=docs%2Fsource%2Fnotebooks%2Fvisualizations.ipynb>`_\n* `Advanced visualizations <https://mybinder.org/v2/gh/lmmentel/mendeleev/master?filepath=docs%2Fsource%2Fnotebooks%2Fadvanced_visualizations.ipynb>`_\n\n\n## Data\n\nA comprehensive list of the available data together with appropriate\nreferences are available in the\n[documentation](http://mendeleev.readthedocs.org). Here the most\nimportant entries are listed:\n\n### Basic properties\n\n- atomic number\n- atomic weight\n- block\n- cas\n- electrons\n- electronic configuration\n- group\n- name\n- neutrons\n- period\n- protons\n- series\n- symbol\n\n### Standardized colors schemes\n\n-   cpk\\_color\n-   jmol\\_color\n-   molcas\\_gv\\_color\n\n### Size related properties\n\n-   atomic radius (Slater, Rahm)\n-   covalent radius (Bragg, Cordero, Pyykko)\n-   ionic radius\n-   metallic radius\n-   van der Waals radius (Alvarez, Batsanov, Bondi, Dreiding, MM3, RT,\n    Truhlar, UFF)\n\n### Electronegativity scales\n\n-   Allen\n-   Allred & Rochow\n-   Cottrell & Sutton\n-   Ghosh\n-   Gordy\n-   Li & Xue\n-   Nagle\n-   Martynov & Batsanov\n-   Mulliken\n-   Pauling\n-   Sanderson\n\n### Descriptive properties\n\n-   discoverers\n-   discovery location\n-   dipole year\n-   description\n-   name origin\n-   sources\n-   uses\n\n### Physical properties\n\n-   boiling point\n-   C\\<sub\\>6\\</sub\\>\n-   density\n-   dipole polarizability\n-   electron affinity\n-   evaporation heat\n-   gas basicity\n-   ionization energies\n-   lattice structure\n-   melting point\n-   oxidation states\n-   proton affinity\n-   specific heat\n-   thermal conductivity\n-   nuclear screening constants (Slater & Clementi)\n\n### Computed properties\n\n-   hardness\n-   softness\n-   electrophilicity\n\n### Isotope properties\n\n-   abundance\n-   g\\_factor\n-   half\\_life\n-   radioactivity\n-   mass\n-   mass number\n-   spin\n-   quadrupole\\_moment\n\n## Getting started\n\nThe simplest way of accessing the element data is by importing elements\ndirectly from the [mendeleev]{.title-ref} package by their symbols. For\nexample consider iron (Fe):\n\n    >>> from mendeleev import Fe\n    >>> Fe.name\n    \'Iron\'\n    >>> Fe.atomic_number\n    26\n    >>> Fe.thermal_conductivity\n    80.4\n\nAnother, more flexible way is through the `element` method that returns\nthe `Element` object:\n\n    >>> from mendeleev import element\n\nThe `element` method accepts unique identifiers: atomic number, atomic\nsymbol or element\\\'s name in english. To retrieve the entries on Silicon\nby symbol type\n\n``` {.sourceCode .python}\n>>> si = element(\'Si\')\n>>> si.name\n\'Silicon\'\n```\n\nSimilarly to access the data by atomic number or element names type\n\n``` {.sourceCode .python}\n>>> al = element(13)\n>>> al.name\n\'Aluminium\'\n>>> o = element(\'Oxygen\')\n>>> o.atomic_number\n8\n```\n\n### Lists of elements\n\nThe `element` method also accepts list or tuple of identifiers and then\nreturns a list of `Element` objects\n\n``` {.sourceCode .python}\n>>> c, h, o = element([\'C\', \'Hydrogen\', 8])\n>>> c.name, h.name, o.name\n(\'Carbon\', \'Hydrogen\', \'Oxygen\')\n```\n\n#### Isotopes\n\nThe `isotopes` attribute returns a list of `Isotope` objects with the\nfollowing attributes per isotope\n\n-   `atomic_number`\n-   `mass`\n-   `abundance`\n-   `mass_number`\n\n``` {.sourceCode .python}\n>>> fe = element(\'Fe\')\n>>> for iso in fe.isotopes:\n...     print(iso)\n 26   55.93494  91.75%    56\n 26   56.93540   2.12%    57\n 26   57.93328   0.28%    58\n 26   53.93961   5.85%    54\n```\n\nThe columns represent the attributes `atomic_number`, `mass`,\n`abundance` and `mass_number` respectively.\n\n### Accesing data tables and the database\n\n[mendeleev](http://mendeleev.readthedocs.org) offers also methods for\naccessing whole tables of data, e.g. table with the data on all isotopes\nand methods for interacting directly with the database engine, for more\ndetails see the [API\ndocumentation](https://mendeleev.readthedocs.io/en/stable/code.html#accessing-data)\nand [this\ntutorial](https://mendeleev.readthedocs.io/en/stable/notebooks/02_tables.html).\n\n### Command line interface utility\n\nFor those who work in the terminal there is a simple command line\ninterface (CLI) for printing the information about a given element. The\nscript name is [element.py]{.title-ref} and it accepts either the symbol\nor name of the element or it\\\'s atomic number as an argument and prints\nthe data about it. For example, to print the properties of silicon type\n\n``` {.sourceCode .bash}\n$ element.py Si\n                            _  _  _  _      _\n                          _(_)(_)(_)(_)_   (_)\n                         (_)          (_)_  _\n                         (_)_  _  _  _  (_)(_)\n                           (_)(_)(_)(_)_   (_)\n                          _           (_)  (_)\n                         (_)_  _  _  _(_)_ (_) _\n                           (_)(_)(_)(_) (_)(_)(_)\n\n\n\nDescription\n===========\n\n  Metalloid element belonging to group 14 of the periodic table. It is\n  the second most abundant element in the Earth\'s crust, making up 25.7%\n  of it by weight. Chemically less reactive than carbon. First\n  identified by Lavoisier in 1787 and first isolated in 1823 by\n  Berzelius.\n\nSources\n=======\n\n  Makes up major portion of clay, granite, quartz (SiO2), and sand.\n  Commercial production depends on a reaction between sand (SiO2) and\n  carbon at a temperature of around 2200 °C.\n\nUses\n====\n\n  Used in glass as silicon dioxide (SiO2). Silicon carbide (SiC) is one\n  of the hardest substances known and used in polishing. Also the\n  crystalline form is used in semiconductors.\n\nProperties\n==========\n\nAbundance crust                                         282000\nAbundance sea                                              2.2\nAnnotation\nAtomic number                                               14\nAtomic radius                                              132\nAtomic radius rahm                                         232\nAtomic volume                                             12.1\nAtomic weight                                           28.085\nAtomic weight uncertainty                                  NaN\nBlock                                                        p\nBoiling point                                             2628\nC6                                                         305\nC6 gb                                                      308\nCas                                                  7440-21-3\nCovalent radius bragg                                      117\nCovalent radius cordero                                    111\nCovalent radius pyykko                                     116\nCovalent radius pyykko double                              107\nCovalent radius pyykko triple                              102\nCovalent radius slater                                     110\nCpk color                                              #daa520\nDensity                                                   2.33\nDipole polarizability                                    37.31\nDiscoverers                                     Jöns Berzelius\nDiscovery location                                      Sweden\nDiscovery year                                            1824\nElectron affinity                                      1.38952\nElectronic configuration                          [Ne] 3s2 3p2\nEn allen                                                 11.33\nEn ghosh                                              0.178503\nEn pauling                                                 1.9\nEvaporation heat                                           383\nFusion heat                                               50.6\nGas basicity                                             814.1\nGeochemical class                                        major\nGoldschmidt class                                    litophile\nGroup id                                                    14\nHeat of formation                                          450\nIs monoisotopic                                           None\nIs radioactive                                           False\nJmol color                                             #f0c8a0\nLattice constant                                          5.43\nLattice structure                                          DIA\nMelting point                                             1683\nMetallic radius                                            117\nMetallic radius c12                                        138\nMolcas gv color                                        #f0c8a0\nName                                                   Silicon\nName origin                    Latin: silex, silicus, (flint).\nPeriod                                                       3\nProton affinity                                            837\nSeries id                                                    5\nSpecific heat                                            0.703\nSymbol                                                      Si\nThermal conductivity                                       149\nVdw radius                                                 210\nVdw radius alvarez                                         219\nVdw radius batsanov                                        210\nVdw radius bondi                                           210\nVdw radius dreiding                                        427\nVdw radius mm3                                             229\nVdw radius rt                                              NaN\nVdw radius truhlar                                         NaN\nVdw radius uff                                           429.5\n```\n\n## Contributing\n\n`mendeleev` is free and always will be. It is developed and maintained by developers in an Open Source manner.\n**Any contributions are welcome**. You could help by writing documentation, pull-requests, report issues or suggesting new features.\n\n### [Issues](https://github.com/lmmentel/mendeleev/issues)\n\nFeel free to submit [issues](https://github.com/lmmentel/mendeleev/issues) regarding:\n\n-   data updates and recommendations\n-   enhancement requests and new useful features\n-   code bugs\n-   data or citation inconsistencies or errors\n\n### [Pull requests](https://github.com/lmmentel/mendeleev/pulls)\n\n-   before stating to work on your pull request please [submit an\n    issue](https://github.com/lmmentel/mendeleev/issues) first\n-   fork the repo on [github](https://github.com/lmmentel/mendeleev)\n-   clone the project to your own machine\n-   commit changes to your own branch\n-   push your work back up to your fork\n-   submit a [pull request](https://github.com/lmmentel/mendeleev/pulls)\n    so that your changes can be reviewed\n\n## License\n\nThe packegs is release under the persmissive MIT license, see [LICENSE](https://github.com/lmmentel/mendeleev/blob/master/LICENSE)\n\n## Citing\n\nIf you use [mendeleev](http://mendeleev.readthedocs.org) in a scientific\npublication, please consider citing the software as\n\n| \xa0\xa0\xa0L. M. Mentel, *mendeleev* - A Python resource for properties of\n  chemical elements, ions and isotopes. , 2014\\-- . Available at:\n  <https://github.com/lmmentel/mendeleev>.\n\nHere\\\'s the reference in the\n[BibLaTeX](https://www.ctan.org/pkg/biblatex?lang=en) format\n\n``` {.sourceCode .latex}\n@software{mendeleev2014,\n   author = {Mentel, Łukasz},\n   title = {{mendeleev} -- A Python resource for properties of chemical elements, ions and isotopes},\n   url = {https://github.com/lmmentel/mendeleev},\n   version = {0.8.0},\n   date = {2014--},\n```\n\n> }\n\nor the older [BibTeX](http://www.bibtex.org/) format\n\n``` {.sourceCode .latex}\n@misc{mendeleev2014,\n   auhor = {Mentel, Łukasz},\n   title = {mendeleev} -- A Python resource for properties of chemical elements, ions and isotopes, ver. 0.8.0},\n   howpublished = {\\url{https://github.com/lmmentel/mendeleev}},\n   year  = {2014--},\n}\n```\n\n## Funding\n\nThis project was supported by the RCN (The Research Council of Norway)\nproject number 239193.\n',
    'author': 'Lukasz Mentel',
    'author_email': 'lmmentel@gmail.com',
    'maintainer': 'Lukasz Mentel',
    'maintainer_email': 'lmmentel@gmail.com',
    'url': 'https://github.com/lmmentel/mendeleev',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<4.0',
}


setup(**setup_kwargs)



from typing import List, Union

import pandas as pd
from sqlalchemy.dialects import sqlite

from mendeleev import element
from mendeleev import __version__ as version

from .db import get_engine, get_session
from .models import Element, IonizationEnergy


def get_zeff(an, method="slater") -> float:
    "A helper function to calculate the effective nuclear charge"

    e = element(an)
    return e.zeff(method=method)


def fetch_table(table: str, **kwargs) -> pd.DataFrame:
    """
    Return a table from the database as :py:class:`pandas.DataFrame`

    Args:
      table: Name of the table from the database
      kwargs: A dictionary of keyword arguments to pass to the :py:func:`pandas.read_qsl`

    Returns:
      df (pandas.DataFrame): Pandas DataFrame with the contents of the table

    Example:
        >>> from mendeleev.fetch import fetch_table
        >>> df = fetch_table('elements')
        >>> type(df)
        pandas.core.frame.DataFrame

    """

    tables = {
        "elements",
        "groups",
        "isotopes",
        "ionicradii",
        "ionizationenergies",
        "oxidationstates",
        "screeningconstants",
        "series",
    }

    if table not in tables:
        raise ValueError(
            f"Table '{table}' not found, available tables are: {', '.join(sorted(tables))}"
        )

    engine = get_engine()
    return pd.read_sql(table, engine, **kwargs)


def fetch_electronegativities(scales: List[str] = None) -> pd.DataFrame:
    """
    Fetch electronegativity scales for all elements as :py:class:`pandas.DataFrame`

    Args:
        scales: list of scale names, defaults to all available scales

    Returns:
      df (pandas.DataFrame): Pandas DataFrame with the contents of the table
    """

    session = get_session()
    engine = get_engine()

    query = session.query(Element.atomic_number).order_by("atomic_number")
    df = pd.read_sql_query(query.statement.compile(dialect=sqlite.dialect()), engine)

    scales = [
        "allen",
        "allred-rochow",
        "cottrell-sutton",
        "ghosh",
        "gordy",
        "li-xue",
        "martynov-batsanov",
        "mulliken",
        "nagle",
        "pauling",
        "sanderson",
    ]

    for scale in scales:
        scale_name = "-".join(s.capitalize() for s in scale.split("-"))
        df.loc[:, scale_name] = [
            element(int(row.atomic_number)).electronegativity(scale=scale)
            for _, row in df.iterrows()
        ]
    return df.set_index("atomic_number")


def fetch_ionization_energies(degree: Union[List[int], int] = 1) -> pd.DataFrame:
    """
    Fetch a :py:class:`pandas.DataFrame` with ionization energies for all elements
    indexed by atomic number.

    Args:
        degree: Degree of ionization, either as int or a list of ints. If a list is
            passed then the output will contain ionization energies corresponding
            to particalr degrees in columns.

    Returns:
        df (pandas.DataFrame): ionization energies, indexed by atomic number
    """

    # validate degree
    if isinstance(degree, (list, tuple, set)):
        if not all(isinstance(d, int) for d in degree) or any(d <= 0 for d in degree):
            raise ValueError("degree should be a list of positive ints")
    elif isinstance(degree, int):
        if degree <= 0:
            raise ValueError(f"degree should be positive")
        degree = [degree]
    else:
        raise ValueError(
            f"degree should be either a positive int or a collection of positive ints, got: {degree}"
        )

    session = get_session()
    engine = get_engine()

    query = session.query(Element.atomic_number).order_by("atomic_number")
    df = pd.read_sql_query(query.statement.compile(dialect=sqlite.dialect()), engine)

    for d in degree:
        query = session.query(IonizationEnergy).filter(IonizationEnergy.degree == d)
        energies = pd.read_sql_query(
            query.statement.compile(dialect=sqlite.dialect()), engine
        )

        df = pd.merge(
            df,
            energies.loc[:, ["atomic_number", "energy"]].rename(
                columns={"energy": "IE{0:d}".format(d)}
            ),
            on="atomic_number",
            how="left",
        )

    return df.set_index("atomic_number")


def fetch_neutral_data() -> pd.DataFrame:
    """
    Get extensive set of data from multiple database tables as pandas.DataFrame
    """

    elements = fetch_table("elements")
    series = fetch_table("series")
    groups = fetch_table("groups")

    elements = pd.merge(
        elements,
        series,
        left_on="series_id",
        right_on="id",
        how="left",
        suffixes=("", "_series"),
    )
    elements = pd.merge(
        elements,
        groups,
        left_on="group_id",
        right_on="group_id",
        how="left",
        suffixes=("", "_group"),
    )

    elements.rename(columns={"color": "series_colors"}, inplace=True)

    for attr in ["hardness", "softness"]:
        elements[attr] = [
            getattr(element(row.symbol), attr)() for _, row in elements.iterrows()
        ]

    elements["mass"] = [
        element(row.symbol).mass_str() for _, row in elements.iterrows()
    ]

    elements.loc[:, "zeff_slater"] = elements.apply(
        lambda x: get_zeff(x["atomic_number"], method="slater"), axis=1
    )
    elements.loc[:, "zeff_clementi"] = elements.apply(
        lambda x: get_zeff(x["atomic_number"], method="clementi"), axis=1
    )

    ens = fetch_electronegativities()
    elements = pd.merge(elements, ens.reset_index(), on="atomic_number", how="left")

    ies = fetch_ionization_energies(degree=1)
    elements = pd.merge(elements, ies.reset_index(), on="atomic_number", how="left")

    return elements


def fetch_ionic_radii(radius: str = "ionic_radius") -> pd.DataFrame:
    """
    Fetch a pandas DataFrame with ionic radii for all the elements.

    Args:
        radius: The radius to be returned either `ionic_radius` or `crystal_radius`

    Returns:
        df (pandas.DataFrame): a table with atomic numbers, symbols and ionic radii for all
            coordination numbers
    """

    if radius not in ["ionic_radius", "crystal_radius"]:
        raise ValueError(
            "radius '{radius}', not found, available radii are: 'ionic_radius', 'crystal_radius'"
        )

    ir = fetch_table("ionicradii")
    return ir.pivot_table(
        columns="coordination", values=radius, index=["atomic_number", "charge"]
    )


def add_plot_columns(elements: pd.DataFrame) -> pd.DataFrame:
    """
    Add columns needed for the creating the plots

    Args:
        elements: pd.DataFrame
    """

    mask = elements["group_id"].notnull()

    elements.loc[mask, "x"] = elements.loc[mask, "group_id"].astype(int)
    elements.loc[:, "y"] = elements.loc[:, "period"].astype(int)

    elements.loc[mask, "group_name"] = (
        elements.loc[mask, "group_id"].astype(int).astype(str)
    )
    elements.loc[~mask, "group_name"] = "f block"

    for period in [6, 7]:
        mask = (elements["block"] == "f") & (elements["period"] == period)
        elements.loc[mask, "x"] = (
            elements.loc[mask, "atomic_number"]
            - elements.loc[mask, "atomic_number"].min()
            + 3
        )
        elements.loc[mask, "y"] = elements.loc[mask, "period"] + 2.5

    # additional columns for positioning of the text

    elements.loc[:, "y_symbol"] = elements["y"] - 0.05
    elements.loc[:, "y_anumber"] = elements["y"] - 0.3
    elements.loc[:, "y_name"] = elements["y"] + 0.18

    return elements


def get_app_data():
    "write a file with the neutral data"

    data = fetch_neutral_data()
    data = add_plot_columns(data)
    fname = "neutral_{0:s}.pkl".format(version)
    data.to_pickle(fname)
    print("wrote file: ", fname)



import pandas as pd
import plotly.graph_objects as go
from plotly.graph_objs.layout import Shape, Annotation
from pandas.api.types import is_float_dtype

from .utils import colormap_column


def create_tile(
    element: pd.Series, color: str, x_offset: float = 0.45, y_offset: float = 0.45
) -> Shape:
    """
    Create tile shape
    """
    return Shape(
        type="rect",
        x0=element["x"] - x_offset,
        y0=element["y"] - y_offset,
        x1=element["x"] + x_offset,
        y1=element["y"] + y_offset,
        line=dict(color=element[color]),
        fillcolor=element[color],
        opacity=0.8,
    )


def create_annotation(
    row: pd.Series,
    attr: str,
    size: int = 10,
    x_offset: float = 0.0,
    y_offset: float = 0.0,
) -> Annotation:
    """
    Create an annotation from pandas series
    """
    return Annotation(
        x=row["x"] + x_offset,
        y=row["y"] + y_offset,
        xref="x",
        yref="y",
        text=row[attr],
        showarrow=False,
        font=dict(family="Roboto", size=size, color="#333333"),
        align="center",
        opacity=0.9,
    )


def periodic_table_plotly(
    elements: pd.DataFrame,
    attribute: str = "atomic_weight",
    cmap: str = "RdBu_r",
    colorby: str = "color",
    decimals: int = 3,
    height: int = 800,
    missing: str = "#ffffff",
    title: str = "Periodic Table",
    wide_layout: bool = False,
    width: int = 1200,
) -> go.Figure:
    """
    Create a periodic table visualization with plotly.Figure

    Args:
        elements : Pandas DataFrame with the data about elements
        attribute : Name of the attribute to be displayed
        cmap : Colormap to use, see matplotlib colormaps
        colorby : Name of the column containig the colors
        decimals : Number of decimals to be displayed in the bottom row of each cell
        height : Height of the figure in pixels
        missing : Hex code of the color to be used for the missing values
        title : Title to appear above the periodic table
        wide_layout: wide layout variant of the periodic table
        width : Width of the figure in pixels
    """
    fig = go.Figure()

    if colorby == "attribute":
        colored = colormap_column(elements, attribute, cmap=cmap, missing=missing)
        elements.loc[:, "attribute_color"] = colored
        colorby = "attribute_color"

    # tiles
    tiles = [create_tile(row, color=colorby) for _, row in elements.iterrows()]
    fig.layout["shapes"] += tuple(tiles)

    # symbols
    fig.layout["annotations"] += tuple(
        elements.apply(create_annotation, axis=1, raw=False, args=("symbol",), size=16)
    )

    # atomic_number
    fig.layout["annotations"] += tuple(
        elements.apply(
            create_annotation, axis=1, raw=False, args=("atomic_number",), y_offset=-0.3
        )
    )

    # name
    fig.layout["annotations"] += tuple(
        elements.apply(
            create_annotation, axis=1, raw=False, args=("name",), y_offset=0.2, size=7
        )
    )

    ac = "display_attribute"
    if is_float_dtype(elements[attribute]):
        elements[ac] = elements[attribute].round(decimals=decimals)
    else:
        elements[ac] = elements[attribute]

    # attribute
    fig.layout["annotations"] += tuple(
        elements.apply(
            create_annotation,
            axis=1,
            raw=False,
            args=(ac,),
            y_offset=0.35,
            size=7,
        )
    )

    if wide_layout:
        tickvals = None
        xrange = [0.5, 32.5]
        yrange = [7.5, 0.5]
    else:
        tickvals = tuple(range(1, 19))
        xrange = [0.5, 18.5]
        yrange = [10.0, 0.5]

    fig.update_layout(
        template="plotly_white",
        height=height,
        width=width,
        title=title,
        xaxis={
            "range": xrange,
            "showgrid": False,
            "fixedrange": True,
            "side": "top",
            "tickvals": tickvals,
        },
        yaxis={
            "range": yrange,
            "showgrid": False,
            "fixedrange": True,
            "tickvals": tuple(range(1, 8)),
            "title": "Period",
        },
    )

    return fig


import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.cm as cmx

from mendeleev.fetch import fetch_table


def add_tile_coordinates(
    df: pd.DataFrame,
    x_coord: str = "group_id",
    y_coord: str = "period",
    include_f_block: bool = True,
    wide_layout: bool = False,
) -> pd.DataFrame:
    """
    Calculate coordinates for the tile centers

    Args:
        df: dataframe
        x_coord: attribute to use as x coordinate
        y_coord: attribute to use as y coordinate
        wide_layout : Show the long version of the periodic table with the f block between
            the s and d blocks
        include_f_block : Show the elements from the f block
    """

    elements = df.copy().sort_values("atomic_number")

    # calculate x and y coordinates of the main group/row elements
    elements.loc[elements[x_coord].notnull(), "x"] = elements.loc[
        elements[x_coord].notnull(), x_coord
    ].astype(int)
    elements.loc[elements[y_coord].notnull(), "y"] = elements.loc[
        elements[y_coord].notnull(), y_coord
    ].astype(int)

    if include_f_block:
        f_mask = elements["block"] == "f"
        elements.loc[f_mask, "x"] = (
            elements[f_mask]
            .groupby("period")
            .apply(lambda x: x["atomic_number"] - x["atomic_number"].min())
            .to_numpy()
            + 3
        )

        if wide_layout:
            mask = (
                (elements["x"] > 2)
                & (elements["block"] != "f")
                & (~elements["symbol"].isin(["La", "Ac"]))
            )
            elements.loc[f_mask, "x"] += 1
            elements.loc[mask, "x"] = elements.loc[mask, "x"] + 14
            elements.loc[f_mask, "y"] = elements.loc[f_mask, "period"]
        else:
            elements.loc[f_mask, "y"] = elements.loc[f_mask, "period"] + 2.5

    return elements


def create_vis_dataframe(
    x_coord: str = "group_id",
    y_coord: str = "period",
    include_f_block: bool = True,
    wide_layout: bool = False,
):
    """
    Base DataFrame for visualizations

    Args:
        x_coord: attribute to use as x coordinate
        y_coord: attribute to use as y coordinate
        include_f_block : show the elements from the f block
        wide_layout : show the long version of the periodic table with the f block between
            the s and d blocks

    """
    elements = fetch_table("elements")
    series = fetch_table("series")
    df = pd.merge(
        elements,
        series,
        left_on="series_id",
        right_on="id",
        suffixes=("", "_series"),
    ).sort_values(by="atomic_number")
    return add_tile_coordinates(
        df,
        x_coord=x_coord,
        y_coord=y_coord,
        include_f_block=include_f_block,
        wide_layout=wide_layout,
    )


def colormap_column(
    elements: pd.DataFrame, column: str, cmap: str = "RdBu_r", missing: str = "#ffffff"
):
    """
    Return a new DataFrame with the same size (and index) as `elements` with a column
    `cmap` containing HEX color mapping from `cmap` colormap.

    Args:
        elements : DataFrame with the data
        column : Name of the column to be color mapped
        cmap : Name of the colormap, see matplotlib.org
        missing : HEX color for the missing values (NaN or None)
    """

    colormap = plt.get_cmap(cmap)
    cnorm = colors.Normalize(vmin=elements[column].min(), vmax=elements[column].max())
    scalarmap = cmx.ScalarMappable(norm=cnorm, cmap=colormap)

    rgba = scalarmap.to_rgba(elements[column])

    colored = pd.Series(
        index=elements.index, data=[colors.rgb2hex(row) for row in rgba]
    )
    mask = elements[column].isnull()
    colored.loc[mask] = missing

    return colored



from collections import OrderedDict

import pandas as pd
from pandas.api.types import is_float_dtype

from bokeh.plotting import figure
from bokeh.models import HoverTool, ColumnDataSource, FixedTicker

from .utils import colormap_column


def periodic_table_bokeh(
    elements: pd.DataFrame,
    attribute: str = "atomic_weight",
    cmap: str = "RdBu_r",
    colorby: str = "color",
    decimals: int = 3,
    height: int = 800,
    missing: str = "#ffffff",
    title: str = "Periodic Table",
    wide_layout: bool = False,
    width: int = 1200,
):
    """
    Use Bokeh to plot the periodic table data contained in the `df`

    Args:
        elements : Pandas DataFrame with the data about elements
        attribute : Name of the attribute to be displayed
        cmap : Colormap to use, see matplotlib colormaps
        colorby : Name of the column containig the colors
        decimals : Number of decimals to be displayed in the bottom row of each cell
        height : Height of the figure in pixels
        missing : Hex code of the color to be used for the missing values
        title : Title to appear above the periodic table
        wide_layout: wide layout variant of the periodic table
        width : Width of the figure in pixels
    """

    # additional columns for positioning of the text

    elements.loc[:, "y_anumber"] = elements["y"] - 0.3
    elements.loc[:, "y_name"] = elements["y"] + 0.2

    if attribute:
        elements.loc[elements[attribute].notnull(), "y_prop"] = (
            elements.loc[elements[attribute].notnull(), "y"] + 0.35
        )
    else:
        elements.loc[:, "y_prop"] = elements["y"] + 0.35

    ac = "display_attribute"
    if is_float_dtype(elements[attribute]):
        elements[ac] = elements[attribute].round(decimals=decimals)
    else:
        elements[ac] = elements[attribute]

    if colorby == "attribute":
        colored = colormap_column(elements, attribute, cmap=cmap, missing=missing)
        elements.loc[:, "attribute_color"] = colored
        colorby = "attribute_color"

    # bokeh configuration

    source = ColumnDataSource(data=elements)

    TOOLS = "hover,save,reset"

    fig = figure(
        title=title,
        tools=TOOLS,
        x_axis_location="above",
        x_range=(elements.x.min() - 0.5, elements.x.max() + 0.5),
        y_range=(elements.y.max() + 0.5, elements.y.min() - 0.5),
        width=width,
        height=height,
        toolbar_location="above",
        toolbar_sticky=False,
    )

    fig.rect("x", "y", 0.9, 0.9, source=source, color=colorby, fill_alpha=0.6)

    # adjust the ticks and axis bounds
    fig.yaxis.bounds = (1, 7)
    fig.axis[1].ticker.num_minor_ticks = 0
    if wide_layout:
        # Turn off tick labels
        fig.axis[0].major_label_text_font_size = "0pt"
        # Turn off tick marks
        fig.axis[0].major_tick_line_color = None  # turn off major ticks
        fig.axis[0].ticker.num_minor_ticks = 0  # turn off minor ticks
    else:
        fig.axis[0].ticker = FixedTicker(ticks=list(range(1, 19)))

    text_props = {
        "source": source,
        "angle": 0,
        "color": "black",
        "text_align": "center",
        "text_baseline": "middle",
    }

    fig.text(
        x="x",
        y="y",
        text="symbol",
        text_font_style="bold",
        text_font_size="15pt",
        **text_props
    )
    fig.text(
        x="x", y="y_anumber", text="atomic_number", text_font_size="9pt", **text_props
    )
    fig.text(x="x", y="y_name", text="name", text_font_size="6pt", **text_props)
    fig.text(x="x", y="y_prop", text=ac, text_font_size="7pt", **text_props)

    fig.grid.grid_line_color = None

    hover = fig.select(dict(type=HoverTool))
    hover.tooltips = OrderedDict(
        [
            ("name", "@name"),
            ("atomic number", "@atomic_number"),
            ("atomic weight", "@atomic_weight"),
            ("EN Pauling", "@en_pauling"),
            ("Electron affinity", "@electron_affinity"),
            ("CPK color", "$color[hex, swatch]:cpk_color"),
            ("Jmol color", "$color[hex, swatch]:jmol_color"),
            ("Molcas GV color", "$color[hex, swatch]:molcas_gv_color"),
            ("electronic configuration", "@electronic_configuration"),
        ]
    )

    return fig



"""
Class abstracting the elctronic configuration
"""

from collections import OrderedDict
import math
import re

import six


ORBITALS = ("s", "p", "d", "f", "g", "h", "i", "j", "k")
SHELLS = ("K", "L", "M", "N", "O", "P", "Q")


def get_l(subshell):
    "Return the orbital angular momentum quantum number for a given subshell"

    if subshell.lower() in ORBITALS:
        return ORBITALS.index(subshell.lower())
    else:
        raise ValueError(
            'wrong subshell label: "{}",'.format(subshell)
            + " should be one of: {}".format(", ".join(ORBITALS))
        )


def subshell_degeneracy(subshell):
    "Return the degeneracy of a given subshell"

    return 2 * get_l(subshell) + 1


def subshell_capacity(subshell):
    """
    Return the subshell capacity (max number of electrons)
    """

    return 2 * subshell_degeneracy(subshell)


def shell_capactity(shell):
    """
    Return the shell capacity (max number of electrons)

    The capacity is :math:`N=2n^{2}`, where :math:`n` is the principal
    quantum number.
    """

    if shell.upper() in SHELLS:
        return 2 * (SHELLS.index(shell.upper()) + 1) ** 2
    else:
        raise ValueError(
            'wrong shell label: "{}",'.format(shell)
            + " should be one of: {}".format(", ".join(SHELLS))
        )


class ElectronicConfiguration(object):
    """Electronic configuration handler"""

    noble = OrderedDict(
        [
            ("He", "1s2"),
            ("Ne", "1s2 2s2 2p6"),
            ("Ar", "1s2 2s2 2p6 3s2 3p6"),
            ("Kr", "1s2 2s2 2p6 3s2 3p6 4s2 3d10 4p6"),
            ("Xe", "1s2 2s2 2p6 3s2 3p6 4s2 3d10 4p6 5s2 4d10 5p6"),
            ("Rn", "1s2 2s2 2p6 3s2 3p6 4s2 3d10 4p6 5s2 4d10 5p6 6s2 4f14 5d10 6p6"),
        ]
    )

    def __init__(self, conf=None, atomre=None, shellre=None):

        self.atomre = atomre
        self.shellre = shellre
        self.conf = conf

    @property
    def conf(self):
        "Return the configuration"
        return self._conf

    @conf.setter
    def conf(self, value):
        "Setter method for initializing the configuration"

        if isinstance(value, six.string_types):
            self.confstr = value
            self.parse(str(value))
        elif isinstance(value, dict):
            self._conf = OrderedDict(
                sorted(value.items(), key=lambda x: (x[0][0] + get_l(x[0][1]), x[0][0]))
            )
        else:
            raise ValueError("<conf> should be str or dict, got {}".format(type(value)))

    @property
    def atomre(self):
        "Regular expression for atomic symbols"
        return self._atomre

    @atomre.setter
    def atomre(self, value):

        if value is None:
            self._atomre = re.compile(r"\[([A-Z][a-z]*)\]")
        else:
            self._atomre = re.compile(value)

    @property
    def shellre(self):
        "Regular expression for the shell"
        return self._shellre

    @shellre.setter
    def shellre(self, value):

        if value is None:
            self._shellre = re.compile(r"(?P<n>\d)(?P<o>[spdfghijk])(?P<e>\d+)?")
        else:
            self._shellre = re.compile(value)

    def parse(self, string):
        """
        Parse a ``string`` with electronic configuration into an
        ``OrderedDict`` representation
        """

        core = {}
        citems = string.split()

        if self.atomre.match(citems[0]):
            symbol = str(self.atomre.match(citems[0]).group(1))
            citems = citems[1:]
            core = [
                self.shellre.match(o).group("n", "o", "e")
                for o in ElectronicConfiguration.noble[symbol].split()
                if self.shellre.match(o)
            ]
            core = OrderedDict(
                [((int(n), o), (int(e) if e is not None else 1)) for (n, o, e) in core]
            )

        valence = [
            self.shellre.match(o).group("n", "o", "e")
            for o in citems
            if self.shellre.match(o)
        ]
        valence = OrderedDict(
            [((int(n), o), (int(e) if e is not None else 1)) for (n, o, e) in valence]
        )

        self._conf = OrderedDict(list(core.items()) + list(valence.items()))

    def get_largest_core(self):
        """
        Find the largest noble gas core possible for the current
        configuration and return the symbol of the corresponding noble
        gas element.
        """

        confset = set(self.conf.items())

        for s, conf in reversed(ElectronicConfiguration.noble.items()):

            ec = ElectronicConfiguration(conf)
            nobleset = set(ec.conf.items())

            ans = confset.issuperset(nobleset)
            if ans:
                return (s, ec)

    def get_valence(self):
        """
        Find the valence configuration i.e. remove the largest noble gas
        core from the current configuration and return the result.
        """

        _, core_conf = self.get_largest_core()

        valence = OrderedDict(set(self.conf.items()) - set(core_conf.conf.items()))

        return ElectronicConfiguration(valence)

    def sort(self, inplace=True):
        "Sort the occupations OD"
        if inplace:
            self.conf = OrderedDict(
                sorted(
                    self.conf.items(), key=lambda x: (x[0][0] + get_l(x[0][1]), x[0][0])
                )
            )
        else:
            return OrderedDict(
                sorted(
                    self.conf.items(), key=lambda x: (x[0][0] + get_l(x[0][1]), x[0][0])
                )
            )

    def electrons_per_shell(self):
        "Return number of electrons per shell as dict"

        return {
            s: sum(v for k, v in self.conf.items() if k[0] == n)
            for n, s in zip(range(1, self.max_n() + 1), SHELLS)
        }

    def shell2int(self):
        "configuration as list of tuples (n, l, e)"
        return [(x[0], get_l(x[1]), x[2]) for x in self.conf]

    def max_n(self):
        "Return the largest value of principal quantum number for the atom"

        return max(shell[0] for shell in self.conf.keys())

    def max_l(self, n):
        """
        Return the largest value of azimutal quantum number for a given
        value of principal quantum number

        Args:
            n : int
                Principal quantum number
        """

        return ORBITALS[max(get_l(x[1]) for x in self.conf.keys() if x[0] == n)]

    def last_subshell(self, wrt="order"):
        "Return the valence shell"

        if wrt.lower() == "order":
            return list(self.conf.items())[-1]
        elif wrt.lower() == "aufbau":
            return sorted(
                self.conf.items(), key=lambda x: (x[0][0] + get_l(x[0][1]), x[0][0])
            )[-1]
        else:
            raise ValueError("wrong <wrt>: {}".format(wrt))

    def nvalence(self, block, method=None):
        "Return the number of valence electrons"

        if block in ["s", "p"]:
            return sum(v for k, v in self.conf.items() if k[0] == self.max_n())
        elif block == "d":
            if method == "simple":
                return 2
            else:
                return (
                    self.conf[(self.max_n(), "s")] + self.conf[(self.max_n() - 1, "d")]
                )
        elif block == "f":
            return 2
        else:
            raise ValueError("wrong block: {}".format(block))

    def ne(self):
        "Return the number of electrons"

        return sum(list(self.conf.values()))

    def unpaired_electrons(self):
        "Return the number of unpaired electrons"

        so = self.spin_occupations()

        return sum(v["unpaired"] for v in so.values())

    def ionize(self, n=1):
        """
        Remove `n` electrons from and return a new `ElectronicConfiguration`
        object"""

        newec = ElectronicConfiguration(str(self.__str__()))

        for _ in range(n):

            if not newec.conf:
                raise ValueError("Cannot ionize further, no more electrons!")

            (n, o), ne = newec.last_subshell()

            if ne > 1:
                newec.conf[(n, o)] = ne - 1
            elif ne == 1:
                newec.conf.pop((n, o))

        return newec

    def spin_occupations(self):
        """
        For each subshell calculate the number of `alpha`, `beta` electrons,
        electron pairs and unpaired electrons
        """

        so = OrderedDict()

        for (n, orb), nele in self.conf.items():

            ssc = subshell_capacity(orb)
            ssd = subshell_degeneracy(orb)

            if nele == ssc:
                so[(n, orb)] = {"pairs": ssd, "alpha": ssd, "beta": ssd, "unpaired": 0}
            else:
                pairs = (nele % ssd) * (nele // ssd)
                alpha = nele - pairs
                beta = nele - alpha
                unpaired = nele - pairs * 2

                so[(n, orb)] = {
                    "pairs": pairs,
                    "alpha": alpha,
                    "beta": beta,
                    "unpaired": unpaired,
                }

        return so

    def spin_only_magnetic_moment(self):
        """
        Return the magnetic moment insluding only spin of the electrons
        and not the angular momentum
        """

        ue = self.unpaired_electrons()

        return math.sqrt(ue * (ue + 2))

    def slater_screening(self, n, o, alle=False):
        """
        Calculate the screening constant using the approach introduced by
        Slater in Slater, J. C. (1930). Atomic Shielding Constants. Physical
        Review, 36(1), 57-64.
        `doi:10.1103/PhysRev.36.57 <http://www.dx.doi.org/10.1103/PhysRev.36.57>`_

        Args:
          n : int
            Principal quantum number
          o : str
            orbtial label, (s, p, d, ...)
          alle : bool
            Use all the valence electrons, i.e. calculate screening for
            an extra electron
        """

        ne = 0 if alle else 1
        coeff = 0.3 if n == 1 else 0.35
        if o in ["s", "p"]:
            # get the number of valence electrons - 1
            vale = float(
                sum(v for k, v in self.conf.items() if k[0] == n and k[1] in ["s", "p"])
                - ne
            )
            n1 = sum(v * 0.85 for k, v in self.conf.items() if k[0] == n - 1)
            n2 = sum(float(v) for k, v in self.conf.items() if k[0] in range(1, n - 1))

        elif o in ["d", "f"]:
            # get the number of valence electrons - 1
            vale = float(
                sum(v for k, v in self.conf.items() if k[0] == n and k[1] == o) - ne
            )

            n1 = sum(float(v) for k, v in self.conf.items() if k[0] == n and k[1] != o)
            n2 = sum(float(v) for k, v in self.conf.items() if k[0] in range(1, n))

        else:
            raise ValueError("wrong valence subshell: ", o)

        return n1 + n2 + vale * coeff

    def to_str(self):
        "Return a string with the configuration"

        return " ".join(
            "{n:d}{s:s}{e:d}".format(n=k[0], s=k[1], e=v) for k, v in self.conf.items()
        )

    def __repr__(self):

        return '<ElectronicConfiguration(conf="{}")>'.format(self.to_str())

    def __str__(self):

        return self.to_str()


def get_spin_strings(sodict, average=True):
    """
    spin strings as numpy arrays

    This should be called for valence only
    """

    alphas = []
    betas = []

    for (n, orb), occ in sodict.items():
        nss = subshell_degeneracy(orb)
        if average:
            alphas.extend([occ["alpha"] / nss] * nss)
            betas.extend([occ["beta"] / nss] * nss)
        else:
            alphas.extend([1.0] * occ["alpha"] + [0.0] * (nss - occ["alpha"]))
            betas.extend([1.0] * occ["beta"] + [0.0] * (nss - occ["beta"]))
    return alphas, betas


def print_spin_occupations(sodict, average=True):
    "Pretty format for the spin occupations"

    alphas = []
    betas = []

    for (n, orb), occ in sodict.items():
        nss = subshell_degeneracy(orb)
        if average:
            fmt = "10.8f"
            a = ", ".join(
                "{0:{fmt}}".format(x, fmt=fmt) for x in [occ["alpha"] / nss] * nss
            )

            b = ", ".join(
                "{0:{fmt}}".format(x, fmt=fmt) for x in [occ["beta"] / nss] * nss
            )

        else:
            a = ", ".join(
                "{0:3.1f}".format(x)
                for x in [1] * occ["alpha"] + [0] * (nss - occ["alpha"])
            )

            b = ", ".join(
                "{0:3.1f}".format(x)
                for x in [1] * occ["beta"] + [0] * (nss - occ["beta"])
            )

        alphas.append(a)
        betas.append(b)
        print("{} alpha: ".format(orb), a)
        print("{} beta : ".format(orb), b)
    return alphas, betas
